# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, QRegExp
from PyQt5.QtGui import QStandardItemModel, QStandardItem, QColor, QRegExpValidator, QFont, QFontMetrics
from PyQt5.QtWidgets import QTableView, QLineEdit, QStyledItemDelegate


class HexValidatorDelegate(QStyledItemDelegate):
    def createEditor(self, widget, option, index):
        if not index.isValid():
            return 0
        editor = QLineEdit(widget)
        validator = QRegExpValidator(QRegExp("[0-9a-fA-F]?[0-9a-fA-F]"))
        editor.setValidator(validator)
        return editor


class HexView(QTableView):

    def __init__(self, context: 'core.context.Context', parent):
        super(__class__, self).__init__(parent)
        self._context = context

        self._tab_id = 0
        self._frame_id = 0
        self._input_text = ""

        chunks = self._chunk_string(self._input_text, 16)

        self._init_item_font()
        self._init_model(chunks)
        self._init_headers(chunks)
        self._init_column_size()
        self._init_listener()

        self.setItemDelegate(HexValidatorDelegate())

    #############################################
    #   Init
    #############################################

    def _init_listener(self):
        """ Initialize change events. """
        self._context.listener().textChanged.connect(self._on_text_change)
        self._context.listener().textSelectionChanged.connect(self._on_selection_change)
        self._context.listener().selectedFrameChanged.connect(self._on_selected_frame_change)

    def _init_item_font(self):
        self._item_font = QFont()
        self._item_font.setFamily('Courier')
        self._item_font.setFixedPitch(True)
        self._item_font.setPointSize(8)

    def _init_column_size(self):
        # BUG: Columns should only have minimal padding.
        # FIX: Columns are resized to content.
        # NOTE:
        #   This is still not optimal but I couldn't figure out how to make columns even smaller.
        #
        #   Here's a list of things I've tried:
        #       * removing QLineEdit from HexValidatorDelegate
        #       * setting font size
        #       * setting section resize and using a bunch of stylesheets
        #       self.setStyleSheet("""
        #           QTableView::item { border: 0px; padding: 0px; margin: 0px; }
        #           QTableWidget::item { border: 0px; padding: 0px; margin: 0px; }
        #           QTableView { border: 0px; padding: 0px; margin: 0px; }
        #       """)
        #       self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        #
        for i in range(0, 16):
            self.resizeColumnToContents(i)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setDefaultAlignment(QtCore.Qt.AlignCenter)
        self.verticalHeader().setFixedWidth(self._init_vertical_header_width())
        self.verticalHeader().setFont(self._item_font)

    def _init_vertical_header_width(self):
        vertical_header_font = QFont()
        vertical_header_font.setFamily('Courier')
        vertical_header_font.setFixedPitch(True)
        vertical_header_font.setPointSize(8)
        vertical_header_width = QFontMetrics(vertical_header_font).width("00000") + 22
        return vertical_header_width

    def _init_headers(self, chunks):
        self._init_vertical_header(chunks, self.model())
        self.horizontalHeader().hide()

    def _init_model(self, chunks):
        model = QStandardItemModel(len(chunks), 1 + 16)
        self._populate_items(chunks, model)
        model.itemChanged.connect(self._on_hex_value_change)
        self.setModel(model)

    def _init_vertical_header(self, chunks, model):
        if len(chunks) > 0:
            model.setVerticalHeaderLabels(["0x{0:04x}".format(i) for i in range(len(chunks))])
        else:
            model.setVerticalHeaderLabels(["0x{0:04x}".format(0)])

    #############################################
    #   Private Interface
    #############################################

    def _populate_items(self, chunks, model):
        model.blockSignals(True)
        if not chunks:
            self._populate_row(0, "", model)
        else:
            for row_index, chunk in enumerate(chunks):
                self._populate_row(row_index, chunk, model)
        model.blockSignals(False)

    def _populate_row(self, row_index, chunk, model):
        hex_chars = self._set_cells(chunk, model, row_index)
        self._lock_empty_cells(hex_chars, model, row_index)
        self._set_text_cell(chunk, model, row_index)

    def _set_text_cell(self, chunk, model, row_index):
        text_item = QStandardItem(chunk)
        text_item.setFont(self._item_font)
        text_item.setFlags(text_item.flags() ^ QtCore.Qt.ItemIsEditable)
        model.setItem(row_index, 16, text_item)

    def _set_cells(self, chunk, model, row_index):
        hex_chars = self._get_hex_view(chunk)
        for col_index, hex_char in enumerate(hex_chars):
            hex_item = QStandardItem(hex_char)
            hex_item.setBackground(QColor("#90FF90"))
            hex_item.setFont(self._item_font)
            hex_item.setTextAlignment(QtCore.Qt.AlignCenter)
            model.setItem(row_index, col_index, hex_item)
        return hex_chars

    def _lock_empty_cells(self, hex_chars, model, row_index):
        for col_index in range(len(hex_chars), 16):
            hex_item = QStandardItem("")
            hex_item.setText("")
            hex_item.setFont(self._item_font)
            hex_item.setFlags(hex_item.flags() ^ QtCore.Qt.ItemIsEditable)
            model.setItem(row_index, col_index, hex_item)

    def _refresh_model(self):
        chunks = self._chunk_string(self._input_text, 16)
        model = self.model()
        model.clear()
        self._populate_items(chunks, model)
        self._init_headers(chunks)
        self._init_column_size()

    def _get_hex_text(self, model):
        hex_string = ""
        for row_index in range(0, model.rowCount()):
            for col_index in range(0, 16):
                item = model.data(model.index(row_index, col_index))
                if item:
                    hex_string += item
        return hex_string

    def _set_hex_text(self, item, model):
        hex_text = ""
        for col_index in range(0, 16):
            hex_char = model.data(model.index(item.row(), col_index))
            if hex_char:
                hex_text += hex_char
        model.blockSignals(True)
        model_item = model.item(item.row(), 16)
        model_item.setText(bytes.fromhex(hex_text).decode('utf-8', errors='surrogateescape'))
        model.blockSignals(False)

    def _prefix_hex_value(self, item, model):
        hex_char = model.data(model.index(item.row(), item.column()))
        if len(hex_char) == 1:
            model.blockSignals(True)
            model.item(item.row(), item.column()).setText("0" + hex_char)
            model.blockSignals(False)

    def _chunk_string(self, string: str, length: int) -> List[str]:
        """ Breaks the string into a list of chunks with a specified max length. """
        return [string[0 + i:length + i] for i in range(0, len(string), length)]

    def _get_hex_view(self, string: str) -> List[str]:
        def get_hex_char(str_char):
            """ Returns the hex representation of an utf-character while ignoring surrogate special characters. """
            return "{0:02x}".format(ord(str_char))[:2]

        return [get_hex_char(string[i]) for i in range(0, len(string))]

    def _update_view(self, tab_id: str, frame_id: str, input_text: str):
        if self._frame_id == frame_id and self._input_text != input_text:
            self._tab_id = tab_id
            self._frame_id = frame_id
            self._input_text = input_text
            self.setData(input_text)

    #############################################
    #   Events
    #############################################

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Return:
            self.edit(self.selectionModel().currentIndex())
        super(__class__, self).keyPressEvent(event)

    def _on_hex_value_change(self, item):
        model = self.model()
        self._prefix_hex_value(item, model)
        self._set_hex_text(item, model)
        hex_string = self._get_hex_text(model)
        self.blockSignals(True)
        self._context.listener().textSubmitted.emit(self._tab_id, self._frame_id,
                                                    bytearray.fromhex(hex_string).decode('utf-8',
                                                                                         errors='surrogateescape'))
        self.blockSignals(False)

    def _on_selection_change(self, tab_id: str, frame_id: str, input_text: str):
        self._update_view(tab_id, frame_id, input_text)

    def _on_text_change(self, tab_id: str, frame_id: str, input_text: str):
        self._update_view(tab_id, frame_id, input_text)

    def _on_selected_frame_change(self, tab_id: str, frame_id: str, input_text: str):
        self._frame_id = frame_id
        self._update_view(tab_id, frame_id, input_text)

    #############################################
    #   Public Interface
    #############################################

    def setData(self, input_text: str):
        self._input_text = input_text
        self._refresh_model()



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    This plugin will reformat HTML.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HTML Beautify', "Thomas Engel", ["lxml"], context)

    def run(self, text):
        from lxml import etree, html
        document_root = html.fromstring(text)
        return etree.tostring(document_root, encoding='unicode', pretty_print=True)



import os
import re
import qtawesome

from PyQt5.QtWidgets import QDialog

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigPreviewDialog


class Plugin(ScriptPlugin):
    """
    Opens a dialog to filter text by certain conditions.
    """

    class Option(object):
        Filter_Term = PluginConfig.Option.Label("filter_term", "Filter:")
        Should_Match_Case = PluginConfig.Option.Label("should_match_case", "Match Case")
        Should_Invert_Match = PluginConfig.Option.Label("should_invert_match", "Invert Lines")
        Is_Regex = PluginConfig.Option.Label("is_regex", "Regex")

    class FilterCodec:

        def run(self, config: PluginConfig, text: str):
            lines = []
            filter_term = config.get(Plugin.Option.Filter_Term).value
            is_regex = config.get(Plugin.Option.Is_Regex).value
            should_match_case = config.get(Plugin.Option.Should_Match_Case).value
            for text_line in text.splitlines():
                try:
                    if self._should_filter(text_line, config):
                        if is_regex and should_match_case:
                            match = re.match(filter_term, text_line)
                            lines.append(match.group(0))
                        elif is_regex:
                            match = re.match(filter_term, text_line, flags=re.IGNORECASE)
                            lines.append(match.group(0))
                        else:
                            lines.append(text_line)
                except Exception as e:
                    # Ignore exceptions - most likely an error in the regex filter string
                    pass
            return os.linesep.join(lines)

        def _should_filter(self, text_line: str, config: PluginConfig):
            filter_term = config.get(Plugin.Option.Filter_Term).value
            is_regex = config.get(Plugin.Option.Is_Regex).value
            should_match_case = config.get(Plugin.Option.Should_Match_Case).value
            should_invert_match = config.get(Plugin.Option.Should_Invert_Match).value
            if is_regex and should_match_case:
                result = re.match(filter_term, text_line, flags=re.IGNORECASE) is not None
            elif is_regex:
                result = re.match(filter_term, text_line) is not None
            elif should_match_case:
                result = filter_term in text_line
            else:
                result = filter_term.lower() in text_line.lower()
            if should_invert_match:
                return not result
            else:
                return result


    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Filter Lines', "Thomas Engel", [], context)
        self._context = context
        self._logger = context.logger
        self._codec = Plugin.FilterCodec()
        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.Filter_Term,
            value="",
            description="term to filter by",
            is_required=True
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.Should_Match_Case,
            value=True,
            description="defines whether filter should match case",
            is_required=False
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.Should_Invert_Match,
            value=False,
            description="defines whether filter should invert match",
            is_required=False
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.Is_Regex,
            value=False,
            description="defines whether filter term is a regex",
            is_required=False
        ))
        self._dialog = None
        self._dialog_return_code = QDialog.Accepted

    def select(self, text: str):
        if not self._dialog:
            self._dialog = PluginConfigPreviewDialog(self._context, self.config.clone(), "Filter Lines",
                                                     self._codec.run, qtawesome.icon("fa.filter"))

        self._dialog.setInput(text)
        self._dialog_return_code = self._dialog.exec_()

        if self._dialog_return_code != QDialog.Accepted:
            # User clicked the Cancel-Button.
            raise AbortedException("Aborted")

        self.config.update(self._dialog.config)
        return self.run(text)

    def title(self):
        return "Filter lines by '{}' using {}".format(
            self.config.get(Plugin.Option.Filter_Term).value, self._getOptionAsHumanReadableString())

    def _getOptionAsHumanReadableString(self):
        options = []
        if self.config.get(Plugin.Option.Should_Match_Case).value:
            options.append('Match Case')
        else:
            options.append('Ignore Case')
        if self.config.get(Plugin.Option.Is_Regex).value:
            options.append('Regular Expression')
        if self.config.get(Plugin.Option.Should_Invert_Match).value:
            options.append('Invert Match')

        return self._join_options_as_human_readable_string(options)

    def run(self, text: str):
        return self._codec.run(self.config, text)



import qtawesome

from PyQt5.QtWidgets import QDialog
from lxml.etree import XMLSyntaxError

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigPreviewDialog


class Plugin(ScriptPlugin):
	"""
	Opens a dialog to filter xml text by certain xpath expression.

	Example:

		Input:
			<a><b>text</b></a>

		Expression:
			//b

		Output:
			<b>text</b>

	"""

	class Option(object):
		Expression = PluginConfig.Option.Label("xpath_expression", "XPath:")

	class Codec:

		def run(self, config: PluginConfig, text: str):
			try:
				from lxml import etree
				expression = config.get(Plugin.Option.Expression).value
				root = etree.fromstring(text)
				return "".join(etree.tostring(item).decode('utf-8', errors="surrogateescape")
					for item in root.xpath(expression)
				)
			except XMLSyntaxError as e:
				raise Exception("Error decoding XML!")
			except Exception as e:
				# Ignore exceptions - most likely an error in the xpath expression
				pass

	def __init__(self, context):
		# Name, Author, Dependencies
		super().__init__('XPath', "Thomas Engel", ["lxml"], context)
		self._context = context
		self._logger = context.logger
		self._codec = Plugin.Codec()
		self.config.add(PluginConfig.Option.String(
			label=Plugin.Option.Expression,
			value="",
			description="xpath expression to filter by",
			is_required=True
		))
		self._dialog = None
		self._dialog_return_code = QDialog.Accepted

	def select(self, text: str):
		if not self._dialog:
			self._dialog = PluginConfigPreviewDialog(self._context, self.config.clone(), "XPath",
													 self._codec.run, qtawesome.icon("fa.filter"))

		self._dialog.setInput(text)
		self._dialog_return_code = self._dialog.exec_()

		if self._dialog_return_code != QDialog.Accepted:
			# User clicked the Cancel-Button.
			raise AbortedException("Aborted")

		self.config.update(self._dialog.config)
		return self.run(text)

	def title(self):
		return "Filter by XPath expression '{}'".format(self.config.get(Plugin.Option.Expression).value)

	def run(self, text: str):
		return self._codec.run(self.config, text)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import uuid

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QVBoxLayout

from dpp.core import Context
from dpp.core.plugin import PluginType, AbstractPlugin, NullPlugin, Plugins, PluginBuilder
from dpp.ui import VSpacer
from dpp.ui.codec_frame_header import CodecFrameHeader
from dpp.ui.combo_box_frame import ComboBoxFrame
from dpp.ui.view.plain_view import PlainView
from dpp.ui.widget.collapsible_frame import CollapsibleFrame
from dpp.ui.widget.identify_format_button import IdentifyFormatButton
from dpp.ui.widget.smart_decode_button import SmartDecodeButton
from dpp.ui.widget.status_widget import StatusWidget


class CodecFrame(CollapsibleFrame):

    upButtonClicked = pyqtSignal(str)  # frame_id
    downButtonClicked = pyqtSignal(str)  # frame_id
    configButtonClicked = pyqtSignal(str)  # frame_id
    closeButtonClicked = pyqtSignal(str)  # frame_id
    pluginSelected = pyqtSignal(str, str, 'PyQt_PyObject')  # frame_id, input_text, plugin
    pluginDeselected = pyqtSignal(str)  # frame_id

    def __init__(self, parent, context: Context, tab_id: str, codec_frames, plugins: Plugins, text):
        super(__class__, self).__init__(parent, context, uuid.uuid4().hex)

        self._tab_id = tab_id
        self._codec_frames = codec_frames
        self._context.shortcutUpdated.connect(self._shortcut_updated_event)
        self._plugins = plugins
        self._flash_event = None

        self._status_widget = StatusWidget(self)
        self.addWidget(self._status_widget)
        self.addWidget(self._init_input_frame(text))
        self.addWidget(self._init_button_frame())

        self._init_header()

        self.show()

    def _init_header(self):

        self._header_frame.addWidget(CodecFrameHeader.TitleHeaderItem(self))
        self._header_frame.addWidget(CodecFrameHeader.ContentPreviewHeaderItem(self))

        self._header_frame.addWidget(CollapsibleFrame.HeaderFrame.VSepItem(self))
        self._header_frame.addWidget(CodecFrameHeader.LineCountInfoHeaderItem(self))

        self._header_frame.addWidget(CollapsibleFrame.HeaderFrame.VSepItem(self))
        self._header_frame.addWidget(CodecFrameHeader.ContentLengthInfoHeaderItem(self))
        self._header_frame.addWidget(CollapsibleFrame.HeaderFrame.VSepItem(self))

        def button_clicked_event(event, signal):
            if event.button() == QtCore.Qt.LeftButton:
                signal.emit(self.id())

        up_button_header_item = CodecFrameHeader.UpButtonHeaderItem(self)
        up_button_header_item.clicked.connect(lambda evt: button_clicked_event(evt, self.upButtonClicked))
        self._header_frame.addWidget(up_button_header_item)

        down_button_header_item = CodecFrameHeader.DownButtonHeaderItem(self)
        down_button_header_item.clicked.connect(lambda evt: button_clicked_event(evt, self.downButtonClicked))
        self._header_frame.addWidget(down_button_header_item)

        config_button_header_item = CodecFrameHeader.ConfigButtonHeaderItem(self)
        config_button_header_item.clicked.connect(lambda evt: button_clicked_event(evt, self.configButtonClicked))
        self._header_frame.addWidget(config_button_header_item)

        close_button_header_item = CodecFrameHeader.CloseButtonHeaderItem(self)
        close_button_header_item.clicked.connect(lambda evt: button_clicked_event(evt, self.closeButtonClicked))
        self._header_frame.addWidget(close_button_header_item)

    def _init_input_frame(self, text):
        input_frame = QFrame(self)
        frame_layout = QVBoxLayout()
        self._plain_view_widget = PlainView(self._tab_id, self._frame_id, text, self._context, self)
        frame_layout.addWidget(self._plain_view_widget)
        frame_layout.setContentsMargins(0, 6, 6, 6)
        input_frame.setLayout(frame_layout)
        return input_frame

    def _init_button_frame(self):
        button_frame = QFrame(self)
        button_frame_layout = QVBoxLayout()
        self._combo_box_frame = ComboBoxFrame(self, self._context)
        self._combo_box_frame.titleSelected.connect(lambda: self.pluginDeselected.emit(self.id()))
        self._combo_box_frame.pluginSelected.connect(lambda plugin: self.pluginSelected.emit(self.id(), self.getInputText(), plugin))
        button_frame_layout.addWidget(self._combo_box_frame)
        self._smart_decode_button = SmartDecodeButton(self,
                                                      self._plugins.filter(type=PluginType.DECODER),
                                                      self._plain_view_widget.toPlainText,
                                                      self.selectComboBoxEntryByPlugin,
                                                      self._logger
        )
        button_frame_layout.addWidget(self._smart_decode_button)
        self._identify_format_button = IdentifyFormatButton(self,
                                                            self._plugins.filter(type=PluginType.DECODER) +
                                                            self._plugins.filter(type=PluginType.IDENTIFY),
                                                            self._plain_view_widget.toPlainText,
                                                            self.selectComboBoxEntryByPlugin,
                                                            self._logger)
        button_frame_layout.addWidget(self._identify_format_button)
        button_frame_layout.addWidget(VSpacer(self))
        button_frame.setLayout(button_frame_layout)
        return button_frame

    def _shortcut_updated_event(self, shortcut):
        id = shortcut.id()
        tooltip = self._get_tooltip_by_shortcut(shortcut)
        combo_box_shortcut_map = {
            Context.Shortcut.FOCUS_DECODER: PluginType.DECODER,
            Context.Shortcut.FOCUS_ENCODER: PluginType.ENCODER,
            Context.Shortcut.FOCUS_HASHER: PluginType.HASHER,
            Context.Shortcut.FOCUS_SCRIPT: PluginType.SCRIPT
        }
        if id in combo_box_shortcut_map:
            self._combo_box_frame.setToolTipByPluginType(combo_box_shortcut_map[id], tooltip)
        else:
            return
        self._logger.debug("Updated tooltip within codec-frame for {id} to {tooltip}".format(id=id, tooltip=tooltip))

    def _update_tooltip(self, the_widget, the_shortcut_id):
        tooltip = self._get_tooltip_by_shortcut_id(the_shortcut_id)
        the_widget.setToolTip(tooltip)

    def _get_tooltip_by_shortcut_id(self, the_shortcut_id):
        shortcut = self._context.getShortcutById(the_shortcut_id)
        return self._get_tooltip_by_shortcut(shortcut)

    def _get_tooltip_by_shortcut(self, shortcut):
        return "{description} ({shortcut_key})".format(description=shortcut.name(), shortcut_key=shortcut.key())

    def id(self) -> str:
        return self._frame_id

    def getPlugin(self) -> AbstractPlugin:
        """ Returns the currently selected plugin. Might return a NullPlugin when nothing is selected. """
        if hasattr(self, '_combo_box_frame'): # A bit dirty, but might be called before initialization (see Frame::isConfigurable)
            return self._combo_box_frame.selectedPlugin()
        else:
            return NullPlugin(self._context)

    def status(self):
        return self._status_widget.status()

    def setStatus(self, type, message):
        self._status_widget.setStatus(type, message)

    def flashStatus(self, status, message):
        self._header_frame.indicateError(status is "ERROR")
        self._status_widget.setStatus(status, message)

    def selectComboBoxEntryByPlugin(self, plugin, blockSignals=False):
        self._combo_box_frame.selectItem(plugin.type(), plugin.name(), block_signals=True)
        if not blockSignals:
            self.pluginSelected.emit(self.id(), self.getInputText(), plugin)

    def toggleSearchField(self):
        self._plain_view_widget.toggleSearchField()

    def setInputText(self, text):
        self._plain_view_widget.blockSignals(True)
        self._plain_view_widget.setPlainText(text)
        self._plain_view_widget.blockSignals(False)
        self.header().refresh()

    def getInputText(self) -> str:
        return self._plain_view_widget.toPlainText()

    def getComboBoxes(self):
        return self._combo_box_frame

    def cutSelectedInputText(self):
        self._plain_view_widget.cutSelectedInputText()

    def copySelectedInputText(self):
        self._plain_view_widget.copySelectedInputText()

    def pasteSelectedInputText(self):
        self._plain_view_widget.pasteSelectedInputText()

    def focusInputText(self):
        self._plain_view_widget.setFocus()

    def hasFocus(self):
        return self._plain_view_widget.hasFocus()

    def focusComboBox(self, type):
        self._combo_box_frame.focusType(type)

    def setPlugin(self, plugin: AbstractPlugin, blockSignals=True):
        if plugin:
            self.selectComboBoxEntryByPlugin(plugin, blockSignals=blockSignals)
            self.getPlugin().setup(plugin.config.toDict())
            self.frameChanged.emit(self.id())

    def title(self) -> str:
        """ Returns the title of the current frame which is either the title of the previous plugin or None. """
        if self.hasPreviousFrame():
            return self.getPreviousFrame().getPlugin().title()

    def name(self) -> str:
        """ Returns the name of the current frame which is either the title of the previous plugin or None. """
        if self.hasPreviousFrame():
            return self.getPreviousFrame().getPlugin().name()

    def type(self) -> str:
        """ Returns the type of the current frame which is either the type of the previous plugin or "None". """
        if self.hasPreviousFrame():
            return self.getPreviousFrame().getPlugin().type() or "None"
        return "None"

    def description(self) -> str:
        """
        Returns the description of the current frame which is either the description of the previous plugin or None.
        """
        if self.hasPreviousFrame():
            if not isinstance(self.getPreviousFrame().getPlugin(), NullPlugin):
                return self.getPreviousFrame().getPlugin().__doc__

    def isConfigurable(self):
        """ Checks whether the plugin which computes the input is configurable. """
        if self.hasPreviousFrame():
            plugin = self.getPreviousFrame().getPlugin()
            return plugin and plugin.is_configurable()
        return False

    def getFrameIndex(self) -> int:
        """ Returns the index of the codec frame. """
        return self._codec_frames.getFrameIndex(self._frame_id)

    def getFrame(self) -> 'dpp.ui.codec_frame.CodecFrame':
        """ Returns the current frame if exists, otherwise an exception is thrown. """
        return self._codec_frames.getFrameByIndex(self.getFrameIndex())

    def hasPreviousFrame(self) -> bool:
        """ Checks whether there is a previous frame. Returns either True or False. """
        return self._codec_frames.hasPreviousFrame(self.getFrameIndex())

    def getPreviousFrame(self) -> 'dpp.ui.codec_frame.CodecFrame':
        """ Returns the previous frame if any, otherwise an exception is thrown. """
        return self._codec_frames.getFrameByIndex(self.getFrameIndex() - 1)

    def hasNextFrame(self) -> bool:
        """ Checks whether there is a next frame. Returns either True or False. """
        return self._codec_frames.hasNextFrame(self.getFrameIndex())

    def getNextFrame(self) -> 'dpp.ui.codec_frame.CodecFrame':
        """ Returns the next frame if any, otherwise an exception is thrown. """
        return self._codec_frames.getFrameByIndex(self.getFrameIndex() + 1)

    def fromDict(self, frame_config):
        self.setInputText(frame_config["text"])
        self.flashStatus(frame_config["status"]["type"], frame_config["status"]["message"])
        if frame_config["plugin"]["name"] and frame_config["plugin"]["type"]:
            # Configure plugin if any. Last frame does not have a plugin configured, yet.
            self.selectComboBoxEntryByPlugin(PluginBuilder(self._context).build(frame_config["plugin"]), blockSignals=True)
        self.header().refresh()
        self.setCollapsed(frame_config["is_collapsed"])

    def toDict(self):
        status_type, status_message = self._status_widget.status()
        return {
            "title": self.title(),
            "text": self.getInputText(),
            "is_collapsed": self.isCollapsed(),
            "status": {
                "type": status_type,
                "message": status_message
            },
            "plugin": self._combo_box_frame.selectedPlugin().toDict()
        }



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a hex string to an integer.

    Example:

        Input:
            0x75bcd15

        Output:
            123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HEX (int)', "Thomas Engel", [], context)

    def run(self, text):
        return self._run_lines(text, lambda text_part: str(int(text_part, 16)))

    def can_decode_input(self, input):
        if len(input) % 2 == 0:
            hex = re.search(r'^(0x|0X)[a-fA-F0-9]+$', input)
            if hex:
                return True
        return False


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy

from PyQt5.QtCore import QTimer, pyqtSignal
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QFrame, QComboBox, QVBoxLayout

import dpp
from dpp.core.plugin import NullPlugin
from dpp.core.plugin import PluginType
from dpp.ui.widget.combo_box import ComboBox


class ComboBoxFrame(QFrame):
    titleSelected = pyqtSignal()
    pluginSelected = pyqtSignal('PyQt_PyObject')

    def __init__(self, parent, context):
        super(ComboBoxFrame, self).__init__(parent)
        self._context = context
        self._plugins = context.plugins()
        self._logger = context.logger()
        self._plugin_history = {}
        layout = QVBoxLayout()
        self._decoder_combo = self._init_combo_box("Decode as ...", PluginType.DECODER)
        self._encoder_combo = self._init_combo_box("Encode as ...", PluginType.ENCODER)
        self._hasher_combo = self._init_combo_box("Hash ...", PluginType.HASHER)
        self._script_combo = self._init_combo_box("Script ...", PluginType.SCRIPT)
        self._combo_boxes = {
            PluginType.DECODER: self._decoder_combo,
            PluginType.ENCODER: self._encoder_combo,
            PluginType.HASHER: self._hasher_combo,
            PluginType.SCRIPT: self._script_combo
        }
        layout.addWidget(self._decoder_combo)
        layout.addWidget(self._encoder_combo)
        layout.addWidget(self._hasher_combo)
        layout.addWidget(self._script_combo)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self._combo_box_selection_history = []

    def _init_combo_box(self, title, type):
        combo_box = ComboBox()
        combo_box.setEditable(True)
        combo_box.setInsertPolicy(QComboBox.NoInsert)
        model = QStandardItemModel()

        model.setItem(0, 0, QStandardItem(title))
        # Do not load disabled plugins and plugins with unresolved dependencies into the combobox.
        plugin_list = [plugin for plugin in self._plugins.filter(type=type) if plugin.is_enabled() and not plugin.check_dependencies()]
        for index, plugin in enumerate(plugin_list):
            item = QStandardItem(plugin.name())
            item.setToolTip(plugin.__doc__)
            model.setItem(index + 1, 0, item)
        combo_box.lineEdit().returnPressed.connect(lambda: self._combo_box_enter_pressed_event(type))
        # BUG: combo_box.lineEdit().selectAll when focussing lineEdit will be deselected again by general focus-event.
        # FIX: Delay select-all for a few milliseconds.
        combo_box.focusInEvent = lambda e: QTimer.singleShot(100, lambda: combo_box.lineEdit().selectAll())
        # BUG: combo_box.lineEdit().deselect does not work correctly when loosing focus.
        # FIX: Delay deselect for a few milliseconds.
        combo_box.focusOutEvent = lambda e: QTimer.singleShot(100, lambda: combo_box.lineEdit().deselect())
        combo_box.setModel(model)
        combo_box.setModelColumn(0)
        combo_box.activated.connect(lambda i: self._combo_box_item_selected_event(type, combo_box.currentIndex()))
        return combo_box

    def _combo_box_enter_pressed_event(self, type):
        combo_box = self._combo_boxes[type]
        current_text = combo_box.currentText()
        self._combo_box_selection_history.append([type, combo_box.index()])
        self.selectItem(type, current_text, block_signals=False)

    def _combo_box_item_selected_event(self, type, index):
        self.resetExceptType(type)
        # Remove frames below when title element was selected
        if index is 0:
            self.titleSelected.emit()
            return

        self._combo_box_selection_history.append([type, index])
        plugin = self.getPluginByTypeAndIndex(type, index)
        self.pluginSelected.emit(plugin)

        if plugin.was_aborted():
            # BUG: Item gets selected although dialog was canceled.
            # FIX: Reselect last item prior to current selection.
            self.reselectLastItem(blockSignals=True)
        else:
            # BUG: Item gets deselected when running dialogs.
            # FIX: Reselect Item
            self._reselect_item(index, type)

    def reselectLastItem(self, blockSignals=True):
        # Always reset everything first.
        self.resetAll()
        if len(self._combo_box_selection_history) > 1:
            # Reselect previous item if any.
            type, index = self._combo_box_selection_history[-2]
            self._reselect_item(index, type)

    def _reselect_item(self, index, type, blockSignals=True):
        combo_box = self._combo_boxes[type]
        combo_box.blockSignals(blockSignals)
        combo_box.setCurrentIndex(index)
        combo_box.blockSignals(False)

    def getPluginByTypeAndIndex(self, type, index) -> dpp.core.plugin.AbstractPlugin:
        try:
            combo_box = self._combo_boxes[type]
            name = combo_box.itemText(index)
            plugin = copy.deepcopy(self._plugins.plugin(name, type))
            if plugin in self._plugin_history:
                plugin = self._plugin_history[plugin]
            else:
                self._plugin_history[plugin] = plugin
            return plugin
        except Exception as e:
            self._logger.error("Unexpected error. {}".format(e))
            return NullPlugin(self._context)

    def selectedPlugin(self) -> dpp.core.plugin.AbstractPlugin:
        selected_plugin_types = [plugin_type for plugin_type in self._combo_boxes.keys() if self._combo_boxes[plugin_type].currentIndex() != 0]
        if len(selected_plugin_types) == 1:
            selected_plugin_type = selected_plugin_types[0]
            selected_combo_box = self._combo_boxes[selected_plugin_type]
            selected_index = selected_combo_box.currentIndex()
            return self.getPluginByTypeAndIndex(selected_plugin_type, selected_index)
        else:
            return NullPlugin(self._context)

    def selectItem(self, type, plugin_name, block_signals=False):
        if not type and not plugin_name:
            self.resetAll()
            return

        self.resetExceptType(type)
        combo_box = self._combo_boxes[type]
        for i in range(combo_box.count()):
            if combo_box.itemText(i) == plugin_name:
                combo_box.blockSignals(block_signals)
                combo_box.setCurrentIndex(i)
                combo_box.blockSignals(False)
                break

    def focusType(self, type: str):
        """ Focues the combo-box associated with the specified type (e.g. PluginType.DECODER, ...). """
        self._combo_boxes[type].setFocus()

    def _reset(self, *combo_boxes, **kwargs):
        """ Resets a list of combo-boxes. """
        for combo_box in combo_boxes:
            combo_box.blockSignals(True)
            combo_box.setCurrentIndex(0)
            combo_box.blockSignals(False)

    def resetExceptType(self, type: str):
        """ Resets all combo-boxes except the specified type (e.g. PluginType.DECODER, ...). """
        for plugin_type in self._combo_boxes.keys():
            if plugin_type is not type:
                self._reset(self._combo_boxes[plugin_type])

    def resetAll(self):
        """ Resets all combo-boxes to show the first element. """
        self._logger.debug("ComboBoxFrame:ResetAll")
        self._reset(self._decoder_combo, self._encoder_combo, self._hasher_combo, self._script_combo)

    def setToolTipByPluginType(self, plugin_type: str, tooltip: str):
        """
        Setup's the tooltip of the combo-box associated with the specified plugin-type.
        :param plugin_type: the plugin-type which is associated with the combo-box (e.g. PluginType.DECODER, ...).
        :param tooltip: the tooltip to show.
        """
        if not plugin_type in self._combo_boxes.keys():
            self._logger.error("Unknown plugin type '{}'. Can not set tooltip to {}.".format(plugin_type, tooltip))
            return
        self._combo_boxes[plugin_type].setToolTip(tooltip)



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a Base64 string.

    Example:

        Input:
            YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXoKXsKwISLCpyQlJi8oKT0/wrRgPD58ICwuLTs6XyMrJyp+CjAxMjM0NTY3ODk=

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE64', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        text = self._add_missing_padding(text.encode('utf-8', errors="surrogateescape"))
        return base64.b64decode(text).decode('utf-8', errors="surrogateescape")

    def _add_missing_padding(self, text):
        missing_padding = len(text) % 4
        if missing_padding != 0:
            text += b'=' * (4 - missing_padding)
        return text

    def can_decode_input(self, input):
        if input.startswith("0o") and input[2:].isdigit():
            # This looks more like an octal encoding.
            return False
        if re.search(r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$', input):
            return True
        return False



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a Base32 string.

    Example:

        Input:
            MFRGGZDFMZTWQ2LKNNWG23TPOBYXE43UOV3HO6DZPIFF5Q \\
            VQEERMFJZEEUTC6KBJHU74FNDAHQ7HYIBMFYWTWOS7EMVS \\
            OKT6BIYDCMRTGQ2TMNZYHE======

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE32', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        return base64.b32decode(text.encode('utf-8', errors='surrogateescape')).decode('utf-8', errors='surrogateescape')

    def can_decode_input(self, input):
        if len(input) % 4 == 0:
            if re.search(r'^(?:[A-Z2-7]{8})*(?:[A-Z2-7]{2}={6}|[A-Z2-7]{4}={4}|[A-Z2-7]{5}={3}|[A-Z2-7]{7}=)?$', input):
                return True
        return False



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes bytes using ZLib.

    Example:

        Input:
            [bytes]

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Zlib', "Thomas Engel", ["zlib"], context)

    def run(self, text):
        import zlib
        return zlib.decompress(text.encode('utf-8', errors='surrogateescape')).decode('utf-8', errors='surrogateescape')

    def can_decode_input(self, input):
        if input:
            input_bytes = input.encode('utf-8', errors='surrogateescape')
            if len(input_bytes) > 2:
                return (input_bytes[0] == 0x78) and (input_bytes[1] in (
                    0x01, # No Compression/Low
                    0x9c, # Default Compression
                    0xda  # Best Compression
                ))
        return False



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dpp.ui.widget.hover_label import HoverLabel


class IconLabel(HoverLabel):
    """ A widget for showing an icon. """

    def __init__(self, parent, icon):
        super(IconLabel, self).__init__(parent)
        self._do_repaint = True
        self._icon = icon
        self._last_height = None
        self.setHoverEffect(False)

    def setIcon(self, icon):
        """ Sets the icon. """
        self._do_repaint = True
        self._icon = icon

    def paintEvent(self, event=None):
        """ Paints the specified icon (if any). """
        super(IconLabel, self).paintEvent(event)
        if self._do_repaint or (self._icon is not None and self.height() is not self._last_height):
            pxm = self._icon.pixmap(self.height(), self.height())
            self.setPixmap(pxm)
            self._last_height = self.height()
            self._do_repaint = False



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a string using Base32.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            MFRGGZDFMZTWQ2LKNNWG23TPOBYXE43UOV3HO6DZPIFF5Q \\
            VQEERMFJZEEUTC6KBJHU74FNDAHQ7HYIBMFYWTWOS7EMVS \\
            OKT6BIYDCMRTGQ2TMNZYHE======
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE32', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        return base64.b32encode(text.encode('utf-8', errors='surrogateescape')).decode('utf-8', errors='surrogateescape')



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using MD5.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            4384c8873a173210f11c30d6ae54baec
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('MD5', "Thomas Engel", ["hashlib"], context)

    def run(self, text):
        import hashlib
        return hashlib.md5(text.encode('utf-8', errors='surrogateescape')).hexdigest()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QPushButton, QMenu, QAction

from dpp.core.plugin import DecoderPlugin


class SmartDecodeButton(QFrame):
    """ A button which provides a smart-decode functionality. """

    def __init__(self, parent, plugins: List[DecoderPlugin], get_input_callback, select_decoder_callback, logger):
        super(__class__, self).__init__(parent)
        self._plugins = plugins
        self._get_input = get_input_callback
        self._select_decoder = select_decoder_callback
        self._logger = logger
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 6, 0, 0)
        self._button = self._init_button()
        layout.addWidget(self._button)
        self.setLayout(layout)

    def _init_button(self):
        button = QPushButton("Smart decode")
        menu = QMenu(self)
        menu.aboutToShow.connect(lambda: self._populate_button_menu(self._get_input()))
        button.setMenu(menu)
        return button

    def _can_plugin_decode_input(self, decoder: DecoderPlugin, input: str) -> bool:
        """ Returns whether the decoder can decode the specified input. """
        try:
            # Check whether decoder thinks it can decode the input
            if not decoder.can_decode_input(input):
                return False
            # Check whether decoder actually can decode the input without any error
            decoder.select(input)
            return True
        except:
            return False

    def _get_matching_decoders(self, input) -> List[DecoderPlugin]:
        """ Returns a list of matching decoders for the specified input. """
        if not input:
            return []

        return [plugin for plugin in self._plugins if self._can_plugin_decode_input(plugin, input)]

    def _populate_button_menu(self, input):
        """ Populates the button menu with a list of matching decoders for the specified input. """
        menu = self._button.menu()
        menu.clear()
        decoders = self._get_matching_decoders(input)
        if not decoders:
            action = menu.addAction("No matching decoders found ...")
            action.setEnabled(False)
            return

        for decoder in decoders:
            action = QAction(decoder.title(), self)
            action.triggered.connect(lambda chk, item=decoder: self._select_decoder(item))
            menu.addAction(action)



import qtawesome
from PyQt5.QtWidgets import QDialog

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigDialog


class Plugin(ScriptPlugin):
    """
    Splits and Rejoins a string.

    Example 1:

        Split by character ' ' and join with ''

        Input:
            ab cd ef gh ij kl mn op qr st uv wx yz

        Output:
            abcdefghijklmnopqrstuvwxyz

    Example 2:

        Split by length '2' and join with ' '

        Input:
            abcdefghijklmnopqrstuvwxyz

        Output:
            ab cd ef gh ij kl mn op qr st uv wx yz

    """

    class Option(object):
        SplitText = PluginConfig.Option.Label("split_term", "Split by:")
        SplitByLength = PluginConfig.Option.Label("split_by_length", "Length")
        SplitByChars = PluginConfig.Option.Label("split_by_chars", "Character")
        RejoinWithChars = PluginConfig.Option.Label("rejoin_with_chars", "Rejoin with:")

    class SplitAndRejoinCodec:

        def _chunk_string(self, string, length):
            return [string[0 + i:length + i] for i in range(0, len(string), length)]

        def run(self, config, input):
            if config.get(Plugin.Option.SplitByLength).is_checked:
                input = self._chunk_string(input, int(config.get(Plugin.Option.SplitText).value))
            elif config.get(Plugin.Option.SplitByChars).is_checked:
                input = input.split(config.get(Plugin.Option.SplitText).value)
            return config.get(Plugin.Option.RejoinWithChars).value.join(input)

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Split & Rejoin', "Thomas Engel", [], context)
        self._init_config()
        self._codec = Plugin.SplitAndRejoinCodec()
        self._dialog = None
        self._dialog_return_code = None

    def _init_config(self):
        def _validate_split_text(config: PluginConfig, codec, input):
            print("WHAT")
            if not config.get(Plugin.Option.SplitText).value:
                return "Split by should not be empty."
            if config.get(Plugin.Option.SplitByLength).value:
                try:
                    length = int(config.get(Plugin.Option.SplitText).value)
                    if length <= 0:
                        return "Split by should be greater than 0."
                except:
                    return "Split by should be an integer."

        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.SplitText,
            value=",", # default, since non-empty is not allowed (see validator)
            description="the parameter used for splitting",
            is_required=True
        ), validator=_validate_split_text)
        self.config.add(PluginConfig.Option.Group(
            label=Plugin.Option.SplitByChars,
            value=True,
            description="specifies whether text should be split at chars",
            is_required=False,
            group_name="split_behaviour"
        ))
        self.config.add(PluginConfig.Option.Group(
            label=Plugin.Option.SplitByLength,
            value=False,
            description="specifies whether text should be split at interval",
            is_required=False,
            group_name="split_behaviour"
        ))
        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.RejoinWithChars,
            value="",
            description="the chars used to join the split text",
            is_required=True
        ))

    def title(self):
        if self.config.get(Plugin.Option.SplitByLength).is_checked:
            return "Split by length {} and rejoin with '{}'".format(
                self.config.get(Plugin.Option.SplitText).value,
                self.config.get(Plugin.Option.RejoinWithChars).value
            )
        elif self.config.get(Plugin.Option.SplitByChars).is_checked:
            return "Split by characters '{}' and rejoin with '{}'".format(
                self.config.get(Plugin.Option.SplitText).value,
                self.config.get(Plugin.Option.RejoinWithChars).value
            )
        else:
            self.logger().debug("Invalid option.")
            return "Split and Rejoin"

    def select(self, text: str):
        if not self._dialog:
            try:
                self._dialog = PluginConfigDialog(
                    self._context, self.config.clone(), "Split & Rejoin", self._codec.run, qtawesome.icon("fa.edit"))
            except BaseException as e:
                self._context.logger().exception(e, exc_info=self._context.isDebugModeEnabled())

        self._dialog.setInput(text)
        self._dialog_return_code = self._dialog.exec_()

        if self._dialog_return_code != QDialog.Accepted:
            # User clicked the Cancel-Button.
            raise AbortedException("Aborted")

        self.config.update(self._dialog.config)

        return self.run(text)

    def run(self, text: str):
        if text:
            return self._codec.run(self.config, text)
        return ''



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using PHPass.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            $P$HaRrI8HUeMkKf2xmFE6mUg/NUtBEzp/
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('PHPass', "Thomas Engel", ["passlib"], context)

    def run(self, text):
        from passlib.hash import phpass
        return phpass.encrypt(text.encode('utf-8', errors='surrogateescape'))



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA384.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            4e6b190af95ffaae6b963ec8cf4138aabe0f33b088164afe \\
            c358f47299ca53885a6e086b74ef7981695e9d55bb809abc
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA384', "Thomas Engel", ["hashlib"], context)

    def run(self, text):
        import hashlib
        return hashlib.sha384(text.encode('utf-8', errors='surrogateescape')).hexdigest()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse


class OrderedMultiArgs(argparse.Action):
    """
    Defines an Argparse-Action which allows to specify an argument multiple times while order is preserved.

    Example:

        parser = argparse.ArgumentParser()
        parser.add_argument('--test1', action=CustomAction)
        parser.add_argument('--test2', action=CustomAction)

        parser.parse_args(['--test2', '2', '--test1', '1', '--test2', '3'])
        Namespace(ordered_args=[('test2', ['2']), ('test1', ['1']), ('test2', ['3'])], test1=True, test2=True)

    """

    def __call__(self, parser, args, values, option_string=None):
        if not 'ordered_args' in args:
            setattr(args, 'ordered_args', [])
        previous = args.ordered_args
        if type(values) != list:
            values = [values]
        previous.append((self.dest, values))
        setattr(args, 'ordered_args', previous)
        setattr(args, self.dest, True)



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a binary string.

    Example:

        Input:
            1100001 1100010 1100011 1100100 1100101 \\
            1100110 1100111 1101000 1101001 1101010 \\
            ...
            110100 110101 110110 110111 111000 111001

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BIN (str)', "Thomas Engel", [], context)

    def run(self, text):
        n = int(text.replace(' ', ''), 2)
        return n.to_bytes((n.bit_length() + 7) // 8, 'big').decode("utf-8", "surrogateescape") or '\0'



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import json

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QAction, QFileDialog

from dpp.core import Context
from dpp.core.plugin import PluginType
from dpp.ui import CodecTab
from dpp.ui.codec_frame import CodecFrame
from dpp.ui.dialog.config_dialog import ConfigDialog
from dpp.ui.main_window_tabs_widget import MainWindowTabsWidget


class MainWindowWidget(QWidget):

    def __init__(self, parent, context: 'core.context.Context', input: str = None):
        super(QWidget, self).__init__(parent)
        self._parent = parent
        self._context = context
        self._plugins = context.plugins()
        self._main_window = parent

        self.layout = QVBoxLayout(self)
        self._tabs = MainWindowTabsWidget(self._context, self._plugins)
        self._tabs.tabAdded.connect(self._tab_added_event)
        self._tabs.tabClosed.connect(self._tab_closed_event)
        self.layout.addWidget(self._tabs)
        self.setLayout(self.layout)

        self._init_shortcuts()
        self._tabs.newTab(input)
        self._tabs.setCurrentIndex(0)

    def _init_shortcuts(self):
        main_menu = self._parent.menuBar()
        self._file_menu = self._init_file_menu(main_menu)
        self._edit_menu = self._init_edit_menu(main_menu)
        self._view_menu = self._init_view_menu(main_menu)
        self._select_menu = self._init_select_menu(main_menu)
        self._tab_menu = self._init_tab_menu(main_menu)
        self._help_menu = self._init_help_menu(main_menu)

    def _init_help_menu(self, main_menu):
        help_menu = main_menu.addMenu('&Help')
        keyboard_shortcuts_action = QAction("&Keyboard Shortcuts...", self)
        keyboard_shortcuts_action.triggered.connect(
            lambda: self._show_hidden_dialog(ConfigDialog.TAB_KEYBOARD_SHORTCUTS))
        help_menu.addAction(keyboard_shortcuts_action)
        about_action = QAction("&About", self)
        about_action.triggered.connect(self._show_hidden_dialog)
        help_menu.addAction(about_action)
        return help_menu

    def _init_edit_menu(self, main_menu):
        edit_menu = main_menu.addMenu('&Edit')

        self._register_shortcut(Context.Shortcut.EDIT_CUT,
                                "C&ut",
                                "Ctrl+X",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.cutSelectedInputText()),
                                edit_menu)
        self._register_shortcut(Context.Shortcut.EDIT_COPY,
                                "&Copy",
                                "Ctrl+C",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.copySelectedInputText()),
                                edit_menu)
        self._register_shortcut(Context.Shortcut.EDIT_PASTE,
                                "&Paste",
                                "Ctrl+P",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.pasteSelectedInputText()),
                                edit_menu)
        edit_menu.addSeparator()
        self._register_shortcut(Context.Shortcut.TOGGLE_SEARCH_FIELD,
                                "Toggle &Search Field",
                                "Ctrl+F",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.toggleSearchField()),
                                edit_menu)
        return edit_menu

    def _init_view_menu(self, main_menu):
        view_menu = main_menu.addMenu('&View')
        self._register_shortcut(Context.Shortcut.SELECT_LOG_DOCK,
                                "&Log",
                                "Alt+Shift+L",
                                lambda: self._main_window._toggle_log_dock(),
                                view_menu)
        self._register_shortcut(Context.Shortcut.SELECT_HEX_DOCK,
                                "He&x",
                                "Alt+Shift+X",
                                lambda: self._main_window._toggle_hex_dock(),
                                view_menu)
        return view_menu

    def _init_select_menu(self, main_menu):
        select_menu = main_menu.addMenu('&Select')
        # self._register_shortcut(Context.Shortcut.FOCUS_CODEC_SEARCH,
        #                        "Select Codec Search",
        #                        "Ctrl+Space",
        #                        lambda: self._tabs.focusCodecSearch(),
        #                        select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_DECODER,
                                "Select &Decoder",
                                "Alt+Shift+D",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.focusComboBox(PluginType.DECODER)),
                                select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_ENCODER,
                                "Select &Encoder",
                                "Alt+Shift+E",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.focusComboBox(PluginType.ENCODER)),
                                select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_HASHER,
                                "Select &Hasher",
                                "Alt+Shift+H",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.focusComboBox(PluginType.HASHER)),
                                select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_SCRIPT,
                                "Select &Script",
                                "Alt+Shift+S",
                                lambda: self._call_focused_frame(
                                    lambda focused_frame: focused_frame.focusComboBox(PluginType.SCRIPT)),
                                select_menu)

        select_menu.addSeparator()

        self._register_shortcut(Context.Shortcut.FOCUS_INPUT_TEXT,
                                "Select &Text Field",
                                "Alt+Shift+I",
                                lambda: self._call_focused_frame(lambda focused_frame: focused_frame.focusInputText()),
                                select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_INPUT_TEXT_NEXT,
                                "Select &Next Text Field",
                                "Alt+Down",
                                lambda: self._focus_input_text(lambda frame: frame.getNextFrame()),
                                select_menu)
        self._register_shortcut(Context.Shortcut.FOCUS_INPUT_TEXT_PREVIOUS,
                                "Select &Previous Text Field",
                                "Alt+Up",
                                lambda: self._focus_input_text(lambda frame: frame.getPreviousFrame()),
                                select_menu)
        return select_menu

    def _init_tab_menu(self, main_menu):
        tab_menu = main_menu.addMenu('&Tabs')
        self._register_shortcut(Context.Shortcut.TAB_RENAME,
                                "&Rename Tab",
                                "Alt+Shift+R",
                                lambda: self._tabs.tabBar().renameTab(),
                                tab_menu)
        self._register_shortcut(Context.Shortcut.TAB_DUPLICATE,
                                "&Duplicate Tab",
                                "Ctrl+Shift+D",
                                lambda: self._tabs.duplicateTab(),
                                tab_menu)
        self._register_shortcut(Context.Shortcut.TAB_NEXT,
                                "&Next Tab",
                                "Ctrl+Tab",
                                lambda: self._tabs.nextTab(),
                                tab_menu)
        self._register_shortcut(Context.Shortcut.TAB_PREVIOUS,
                                "&Previous Tab",
                                "Ctrl+Shift+Tab",
                                lambda: self._tabs.previousTab(),
                                tab_menu)

        tab_menu.addSeparator()
        self._tabs_select_action = []

        def select_tab(index):
            return lambda: self._tabs.selectTab(index)

        for tab_num in range(1, 10):
            action = self._register_shortcut(Context.Shortcut.TAB_SELECT_.format(tab_num),
                                             "Select Tab &{}".format(tab_num),
                                             "Alt+{}".format(tab_num),
                                             select_tab(tab_num - 1),
                                             tab_menu)
            action.setVisible(False)
            self._tabs_select_action.append(action)
        return tab_menu

    def _init_file_menu(self, main_menu):
        file_menu = main_menu.addMenu('&File')
        self._register_shortcut(Context.Shortcut.TAB_NEW,
                                "&New Tab", "Ctrl+T",
                                lambda: self._tabs.newTab(),
                                file_menu)
        self._register_shortcut(Context.Shortcut.TAB_CLOSE,
                                "&Close Tab",
                                "Ctrl+W",
                                lambda: self._tabs.closeTab(),
                                file_menu)
        file_menu.addSeparator()
        self._register_shortcut(Context.Shortcut.OPEN_FILE,
                                "&Open File...", "Ctrl+O",
                                lambda: self._open_file(),
                                file_menu)
        self._register_shortcut(Context.Shortcut.SAVE_AS_FILE,
                                "&Save As...", "Ctrl+S",
                                lambda: self._save_as_file(),
                                file_menu)
        file_menu.addSeparator()
        self._register_shortcut(Context.Shortcut.SHOW_PLUGINS,
                                "&Plugins...", "Ctrl+Shift+P",
                                lambda: self._show_plugins_dialog(),
                                file_menu)
        file_menu.addSeparator()
        self._register_shortcut(Context.Shortcut.FILE_EXIT,
                                "E&xit",
                                "Ctrl+Q",
                                lambda: self._parent.close(),
                                file_menu)
        return file_menu

    def _register_shortcut(self, id, text, shortcut_key, callback, menu) -> QAction:
        action = self._context.registerShortcut(id, text, shortcut_key, callback, self)
        menu.addAction(action)
        return action

    def _call_focused_frame(self, callback):
        focused_frame = self._get_focused_frame()
        if focused_frame:
            callback(focused_frame)

    def _get_focused_frame(self) -> CodecFrame:
        return self._tabs.currentWidget().frames().getFocusedFrame()

    def _focus_input_text(self, callback):
        frame = self._get_focused_frame()
        if frame:
            future_frame = callback(frame) or frame
            future_frame.focusInputText()
            # Collapse/Uncollapse frames automatically.
            if self._tabs.currentWidget().frames().getFramesCount() > 2:
                if frame != future_frame:
                    if future_frame.isCollapsed():
                        if future_frame.hasPreviousFrame():
                            # Toggle frame, except the first frame which does not have a header
                            future_frame.toggleCollapsed()
                    if not frame.wasCollapseStateChangedByUser():
                        if frame.hasPreviousFrame():
                            # Collapse frame, except the first frame which does not have a header
                            frame.toggleCollapsed()

            self._tabs.currentWidget().ensureWidgetVisible(future_frame)

    def _show_hidden_dialog(self, tab_select: str = None):
        """ Shows the hidden dialog. """
        hidden_dialog = ConfigDialog(self, self._context, tab_select)
        hidden_dialog.exec_()

    def _open_file(self):
        filename, _ = QFileDialog.getOpenFileName(self, 'Open File')
        if filename:
            try:
                with open(filename) as f:
                    save_file = json.loads(f.read())
                    for tab_config in save_file:
                        index, tab = self.newTab(title=tab_config["name"])
                        frame_index = 0
                        for frame_config in tab_config["frames"]:
                            if frame_index == 0:
                                # New tabs already contain one empty frame
                                frame = tab.frames().getFrames()[0]
                                frame.setInputText(frame_config["text"])
                                frame.setStatus(frame_config["status"]["type"], frame_config["status"]["message"])
                            else:
                                frame = tab.frames().newFrame(frame_config["text"],
                                                              frame_config["title"],
                                                              frame_index,
                                                              frame_config["status"]["type"],
                                                              frame_config["status"]["message"])
                            frame.fromDict(frame_config)
                            frame_index = frame_index + 1
                self._context.logger().info("Successfully loaded {}!".format(filename))
            except Exception as e:
                self._context.logger().error("Unexpected error loading file. {}".format(e))

    def _save_as_file(self):
        filename, _ = QFileDialog.getSaveFileName(self, 'Save As File')
        if filename:
            try:
                self._context.saveAsFile(filename, str(json.dumps(self._tabs.toDict(), default=lambda x: x.__dict__)))
                self._context.logger().info("Successfully saved session in {}!".format(filename))
            except Exception as e:
                self._context.logger().error("Unexpected error saving file. {}".format(e))

    def _show_plugins_dialog(self):
        self._show_hidden_dialog("Plugins")

    def _tab_added_event(self, index, name):
        if 0 <= index < len(self._tabs_select_action):
            self._tabs_select_action[index].setVisible(True)

    def _tab_closed_event(self, index, name):
        for index in range(0, len(self._tabs_select_action)):
            self._tabs_select_action[index].setVisible(index < self._tabs.tabCount())

    def newTab(self, input: str = None, title: str=None) -> (int, CodecTab):
        """
        Opens a new tab and writes input into first codec-frame.
        :param input: The input which should be placed into the first codec-frame.
        :param title: The title of the tab.
        """
        return self._tabs.newTab(input, title)



import re

import qtawesome
from PyQt5.QtWidgets import QDialog

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigDialog


class Plugin(ScriptPlugin):
    """
    Opens a search-and-replace dialog which supports Match-Case and Regular Expressions.
    """

    class Option(object):
        SearchTerm = PluginConfig.Option.Label("search_term", "Search:")
        ReplaceTerm = PluginConfig.Option.Label("replace_term", "Replace:")
        ShouldMatchCase = PluginConfig.Option.Label("should_match_case", "Match Case")
        IsRegex = PluginConfig.Option.Label("is_regex", "Regex")

    class SearchAndReplaceCodec:

        def run(self, config, input):
            is_regex = config.get(Plugin.Option.IsRegex).value
            should_match_case = config.get(Plugin.Option.ShouldMatchCase).value
            search_term = config.get(Plugin.Option.SearchTerm).value
            replace_term = config.get(Plugin.Option.ReplaceTerm).value
            if is_regex and should_match_case:
                return re.sub(search_term, replace_term, input, flags=re.IGNORECASE)
            elif is_regex:
                return re.sub(search_term, replace_term, input)
            elif should_match_case:
                return input.replace(search_term, replace_term)
            else:
                regexp = re.compile(re.escape(search_term), re.IGNORECASE)
                return regexp.sub(search_term, input)

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Search & Replace', "Thomas Engel", [], context)
        self._codec = Plugin.SearchAndReplaceCodec()
        self._init_config()

    def _init_config(self):
        def _validate_search_term(config: PluginConfig, codec, input):
            if not config.get(Plugin.Option.SearchTerm).value:
                return "Search term should not be empty."

        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.SearchTerm,
            value="",
            description="the word or phrase to replace",
            is_required=True
        ), validator=_validate_search_term)
        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.ReplaceTerm,
            value="",
            description="the word or phrase used as replacement",
            is_required=True
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.ShouldMatchCase,
            value=True,
            description="defines whether the search term should match case",
            is_required=False
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.IsRegex,
            value=False,
            description="defines whether the search term is a regular expression",
            is_required=False
        ))
        self._dialog = None
        self._dialog_return_code = QDialog.Accepted

    def select(self, text: str):
        if not self._dialog:
            self._dialog = PluginConfigDialog(self._context, self.config.clone(), "Search & Replace", self._codec.run,
                                              qtawesome.icon("fa.search"))

        self._dialog.setInput(text)
        self._dialog_return_code = self._dialog.exec_()

        if self._dialog_return_code != QDialog.Accepted:
            # User clicked the Cancel-Button.
            raise AbortedException("Aborted")

        self.config.update(self._dialog.config)
        return self.run(text)

    def title(self):
        return "Search and Replace '{}' with '{}' using {}".format(
            self._get_search_term(), self._get_replace_term(), self._get_option_as_human_readable_string())

    def _get_option_as_human_readable_string(self):
        if self._should_match_case():
            return 'Match Case'
        elif self._is_regex():
            return 'Regular Expression'
        else:
            return 'Ignore Case'

    def _get_search_term(self):
        return self.config.get(Plugin.Option.SearchTerm).value

    def _get_replace_term(self):
        return self.config.get(Plugin.Option.ReplaceTerm).value

    def _should_match_case(self):
        return self.config.get(Plugin.Option.ShouldMatchCase).value

    def _is_regex(self):
        return self.config.get(Plugin.Option.IsRegex).value

    def run(self, text: str):
        return self._codec.run(self.config, text)



import math
import string

import qtawesome
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeySequence, QIntValidator
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QVBoxLayout, QLineEdit, QFrame, QPlainTextEdit, QShortcut, \
    QSlider, QHBoxLayout, QPushButton

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig


class Plugin(ScriptPlugin):
    """ Opens a dialog to transform text using caesar-cipher. """

    class Option(object):

        Shift = PluginConfig.Option.Label("shift", "Shift:")

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Caesar Cipher', "Thomas Engel", [], context)

        self.config.add(PluginConfig.Option.Integer(
            label=Plugin.Option.Shift,
            value=0,
            description="integer by which the value of the letters should be shifted.",
            is_required=True,
            range=[0, 26]
        ))
        self._dialog = None
        self._codec = CaesarCipher()

    def select(self, text: str):
        if not self._dialog:
            self._dialog = CaesarCipherDialog(text, self.config.clone(), self._codec)
        else:
            self._dialog.setInput(text)

        if self._dialog.exec_() != QDialog.Accepted:
            # User clicked the Cancel-Button.
            raise AbortedException("Aborted")

        self.config.update(self._dialog.config)
        return self.run(text)

    def title(self):
        return "{} shift {}".format("Caesar Cipher", self.config.get(Plugin.Option.Shift).value)

    def run(self, text: str):
        return self._codec.run(text, self.config.get(Plugin.Option.Shift).value)


class CaesarCipher:
    """
    Caesar cipher and offset calculator for Caesar cipher based on known frequency of english letters.
    Based on code from RobSpectre (see https://github.com/RobSpectre/Caesar-Cipher/)
    """

    def __init__(self):
        # Frequency of letters used in English, taken from Wikipedia.
        # http://en.wikipedia.org/wiki/Letter_frequency
        self.frequency = {
            'a': 0.08167,
            'b': 0.01492,
            'c': 0.02782,
            'd': 0.04253,
            'e': 0.130001,
            'f': 0.02228,
            'g': 0.02015,
            'h': 0.06094,
            'i': 0.06966,
            'j': 0.00153,
            'k': 0.00772,
            'l': 0.04025,
            'm': 0.02406,
            'n': 0.06749,
            'o': 0.07507,
            'p': 0.01929,
            'q': 0.00095,
            'r': 0.05987,
            's': 0.06327,
            't': 0.09056,
            'u': 0.02758,
            'v': 0.00978,
            'w': 0.02360,
            'x': 0.00150,
            'y': 0.01974,
            'z': 0.00074}

    def _calculate_entropy(self, input):
        """
        Calculates the entropy of a string based on known frequency.
        :param input: the input string.
        :return: a negative float with the total entropy of the input (higher is better).
        """
        total = 0
        for char in input:
            if char.isalpha() and char in self.frequency:
                prob = self.frequency[char.lower()]
                total += - math.log(prob) / math.log(2)
        return total

    def run(self, input, offset):
        """
        Applies the caesar cipher.
        :param input: the input string.
        :param offset: integer by which the value of the letters should be shifted.
        :return: String with cipher applied.
        """
        def translate(input, offset, alphabet):
            shifted_alphabet = alphabet[offset:] + alphabet[:offset]
            table = str.maketrans(alphabet, shifted_alphabet)
            return input.translate(table)
        return translate(translate(input, offset, string.ascii_lowercase), offset, string.ascii_uppercase)

    def calculate_offset(self, input):
        """
        Attempts to calculate offset of ciphertext using frequency of letters in English.
        :param input: the input string.
        :return: the most likely offset
        """
        entropy_values = {}
        for i in range(26):
            offset = i * -1
            test_cipher = self.run(input, offset)
            entropy_values[i] = self._calculate_entropy(test_cipher)

        sorted_by_entropy = sorted(entropy_values, key=entropy_values.get)
        offset = sorted_by_entropy[0] * -1
        return 26 + offset


class CaesarCipherDialog(QDialog):

    def __init__(self, input: str, config: PluginConfig, codec):
        super(CaesarCipherDialog, self).__init__()
        self.config = config
        main_layout = QVBoxLayout()
        main_layout.addWidget(self._init_editor_frame())
        main_layout.addWidget(self._init_button_box())
        self.setLayout(main_layout)
        self.setWindowIcon(qtawesome.icon("fa.edit"))
        self.setWindowTitle("Caesar Cipher")
        self._setup_shortcuts()
        self._input = input
        self._text_area.setPlainText(self._input)
        self._codec = codec

    #############################################
    #   Initialize
    #############################################

    def _setup_shortcuts(self):
        ctrl_return_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Return), self)
        ctrl_return_shortcut.activated.connect(self._accept)
        alt_return_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_Return), self)
        alt_return_shortcut.activated.connect(self._accept)
        alt_o_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_O), self)
        alt_o_shortcut.activated.connect(self._accept)

    def _init_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self._accept)
        button_box.rejected.connect(self.reject)
        return button_box

    def _init_editor_frame(self):
        main_frame = QFrame()
        main_frame_layout = QVBoxLayout()

        slider_frame = self._init_slider_frame()
        main_frame_layout.addWidget(slider_frame)

        self._text_area = QPlainTextEdit()
        self._text_area.setReadOnly(True)
        self._text_area.setFixedHeight(126)
        main_frame_layout.addWidget(self._text_area)

        main_frame.setLayout(main_frame_layout)
        return main_frame

    def _init_slider_frame(self):
        slider_frame = QFrame()
        slider_frame_layout = QHBoxLayout()

        self._shift_slider = QSlider(Qt.Horizontal)
        self._shift_slider.setMinimum(0)
        self._shift_slider.setMaximum(26)
        self._shift_slider.setValue(self.config.get(Plugin.Option.Shift).value)
        self._shift_slider.valueChanged.connect(self._shift_slider_changed)
        slider_frame_layout.addWidget(self._shift_slider)

        self._shift_text = QLineEdit()
        self._shift_text.setText(str(self.config.get(Plugin.Option.Shift).value))
        self._shift_text.setFixedWidth(30)
        self._shift_text.setValidator(QIntValidator(0, 26))
        self._shift_text.textChanged.connect(self._shift_text_changed)
        slider_frame_layout.addWidget(self._shift_text)

        self._shift_calculate_button = QPushButton("Calculate")
        self._shift_calculate_button.clicked.connect(self._shift_calculate_button_clicked)
        slider_frame_layout.addWidget(self._shift_calculate_button)

        slider_frame.setLayout(slider_frame_layout)
        return slider_frame

    #############################################
    #   Events
    #############################################

    def _shift_calculate_button_clicked(self):
        offset = self._codec.calculate_offset(self._input)
        self._shift_slider.setSliderPosition(offset)

    def _shift_slider_changed(self, shift):
        if not shift:
            shift = 0
        self._shift_changed(shift)

    def _shift_text_changed(self, shift):
        if not shift:
            shift = 0
        self._shift_changed(int(shift))

    def _shift_changed(self, shift):
        self._shift_text.blockSignals(True)
        self._shift_slider.blockSignals(True)
        self._shift_slider.setValue(shift)
        self._shift_text.setText(str(shift))
        self._text_area.setPlainText(self._codec.run(self._input, shift))
        self._shift_slider.blockSignals(False)
        self._shift_text.blockSignals(False)

    #############################################
    #   Private interface
    #############################################

    def _get_shift(self):
        return self._shift_slider.value()

    def _accept(self):
        self.config.update({Plugin.Option.Shift.key: self._get_shift()})
        self.accept()

    #############################################
    #   Public interface
    #############################################

    def setInput(self, input) -> str:
        self._input = input
        self._shift_changed(self._get_shift())



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import qtawesome
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QPainter, QColor
from PyQt5.QtWidgets import QFrame, QVBoxLayout, QHBoxLayout, QBoxLayout, QSizePolicy, QWidget

from dpp.core import Context
from dpp.ui import HSpacer
from dpp.ui.widget.frame import Frame
from dpp.ui.widget.separater_widget import VSep


class CollapsibleFrame(Frame):
    """
    Frame with ability to un-/collapse content via clickable arrow-handles.
    """

    arrowClicked = pyqtSignal()
    upButtonClicked = pyqtSignal()
    downButtonClicked = pyqtSignal()
    configButtonClicked = pyqtSignal()
    closeButtonClicked = pyqtSignal()

    def __init__(self, parent, context: Context, frame_id: str):
        """
        Initializes the collapsible frame.
        :param parent: the parent widget of this frame
        """
        super(__class__, self).__init__(parent, context, frame_id)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self._is_collasped = False
        self._was_collapse_state_changed_by_user = False
        self.arrowClicked.connect(self._arrow_clicked_event)
        self._header_frame = self._init_header_frame()
        self._content = self._init_content(QBoxLayout.LeftToRight, self._is_collasped)
        self._init_frame_style()

        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignTop)

        self._layout.addWidget(self._header_frame)
        self._layout.addWidget(self._content)

        self.setLayout(self._layout)

    def _init_frame_style(self):
        background_color = self.palette().color(self.backgroundRole())
        self.setStyleSheet(
            "CollapsibleFrame { border:1px solid " + background_color.lighter(
                90).name() + "; background:" + background_color.lighter(99).name() + "; }")

    def _init_header_frame(self):
        header_frame = CollapsibleFrame.HeaderFrame(self)
        header_frame.arrowClicked.connect(self.arrowClicked.emit)
        return header_frame

    def _init_content(self, direction, collapsed):
        content = QFrame(self)
        content_layout = QBoxLayout(direction)
        content_layout.setAlignment(Qt.AlignTop)
        content.setLayout(content_layout)
        content.setVisible(not collapsed)
        return content

    def _arrow_clicked_event(self):
        self.toggleCollapsed()
        self._was_collapse_state_changed_by_user = True

    def addWidget(self, widget):
        """
        Adds an widget to the content.
        :param widget: the widget to be added.
        """
        self._content.layout().addWidget(widget)

    def setContentDirection(self, direction: int):
        """ Sets the direction of the content (e.g. LeftToRight, RightToLeft, TopToBottom, BottomToTop). """
        self._content.layout().setDirection(direction)

    def toggleCollapsed(self):
        """ Toggles collapsing of the frame. """
        self._content.setVisible(self._is_collasped)
        self._is_collasped = not self._is_collasped
        self._header_frame._arrow.setArrow(int(self._is_collasped))

    def setCollapsed(self, status):
        """ Un-/collapses the frame. """
        self._content.setVisible(not status)
        self._is_collasped = status
        self._header_frame._arrow.setArrow(int(self._is_collasped))

    def isCollapsed(self) -> bool:
        return self._is_collasped

    def wasCollapseStateChangedByUser(self):
        """ Returns whether the collapse state was changed by the user. """
        return self._was_collapse_state_changed_by_user

    def header(self) -> 'dpp.ui.collapsible_frame.CollapsibleFrame.HeaderFrame':
        return self._header_frame

    def indicateError(self, status: bool):
        """ Indicates that there was an error during encoding/decoding/hashing/scripting. """
        self._header_frame.indicateError(status)

    def getContentHeight(self) -> int:
        """ Returns the height of the content when uncollapsed. """
        return self._content.height()

    def setContentHeight(self, height: int):
        """ Sets the height of the content when uncollapsed. """
        self._content.setFixedHeight(height)

    class HeaderFrame(QFrame):
        """ Clickable frame with title. """

        class AbstractHeaderFrameItem(QFrame):

            def __init__(self, parent):
                super(__class__, self).__init__(parent)
                self.setContentsMargins(0, 0, 0, 0)
                self.setLayout(QHBoxLayout())
                self.layout().setContentsMargins(0, 0, 0, 0)
                self.layout().setSpacing(0)

            def clearLayout(self):
                while self.layout().count():
                    child = self.layout().takeAt(0)
                    if child.widget():
                        child.widget().deleteLater()

            def setCentralWidget(self, widget: QWidget):
                self.clearLayout()
                self.layout().addWidget(widget)

            def refresh(self):
                pass

        class VSepItem(AbstractHeaderFrameItem):

            def __init__(self, parent):
                super(__class__, self).__init__(parent)
                self.setCentralWidget(VSep(self))

        class HSpacerItem(AbstractHeaderFrameItem):

            def __init__(self, parent):
                super(__class__, self).__init__(parent)
                self.setCentralWidget(HSpacer(self))

        arrowClicked = pyqtSignal()
        upButtonClicked = pyqtSignal()
        downButtonClicked = pyqtSignal()
        configButtonClicked = pyqtSignal()
        closeButtonClicked = pyqtSignal()

        def __init__(self, parent: 'ui.widget.CollapsibleFrame'):
            super(__class__, self).__init__(parent)

            self._parent = parent
            self.setMinimumHeight(26)
            self.setMaximumHeight(26)
            self.indicateError(False)

            self._hlayout = QHBoxLayout()
            self._hlayout.setContentsMargins(0, 0, 0, 0)
            self._hlayout.setSpacing(0)

            self._arrow = None

            self._hlayout.addWidget(self._init_header_frame(parent.isCollapsed()))
            self.setLayout(self._hlayout)

        def _init_header_frame(self, collapsed: bool):
            self._frm_header = QFrame(self)
            self._frm_header.setMinimumHeight(24)
            self._frm_header.setMaximumHeight(24)
            frm_header_layout = QHBoxLayout(self)
            frm_header_layout.setContentsMargins(0, 0, 0, 0)
            frm_header_layout.addWidget(self._init_arrow(collapsed))
            background_color = self._frm_header.palette().color(self._frm_header.backgroundRole())
            self._frm_header.setStyleSheet("QFrame { border:0px; background:" + background_color.darker(96).name() + "; }")
            self._frm_header.setLayout(frm_header_layout)
            return self._frm_header

        def _init_arrow(self, collapsed):
            self._arrow = CollapsibleFrame.Arrow(self, collapsed=collapsed)
            self._arrow.setMinimumWidth(24)
            self._arrow.setMaximumWidth(24)
            self._arrow.setStyleSheet("border:0px")
            return self._arrow

        def addWidget(self, header_item: AbstractHeaderFrameItem):
            self._frm_header.layout().addWidget(header_item)

        def refresh(self):
            """ Refreshes the header model (e.g. line-count, character-count, etc.). """
            for i in range(self._frm_header.layout().count()):
                widget = self._frm_header.layout().itemAt(i).widget()
                if isinstance(widget, CollapsibleFrame.HeaderFrame.AbstractHeaderFrameItem):
                    widget.refresh()

        def indicateError(self, status: bool):
            """ Indicates an error by painting the title-border red. Otherweise black. """
            if status:
                self.setStyleSheet("QFrame { border:1px solid red; }")
            else:
                self.setStyleSheet("QFrame { border:1px solid rgb(41, 41, 41); }")

        def mouseReleaseEvent(self, event):
            if self.underMouse() and event.button() == QtCore.Qt.LeftButton:
                # The arrow (or something non-button like) was clicked
                self.arrowClicked.emit()
            return super(CollapsibleFrame.HeaderFrame, self).mousePressEvent(event)

    class Arrow(QFrame):

        def __init__(self, parent=None, collapsed=False):
            super(__class__, self).__init__(parent=parent)

            self.setMaximumSize(24, 24)

            # horizontal == 0
            self._arrow_horizontal = (QtCore.QPointF(7.0, 8.0), QtCore.QPointF(17.0, 8.0), QtCore.QPointF(12.0, 13.0))
            # vertical == 1
            self._arrow_vertical = (QtCore.QPointF(8.0, 7.0), QtCore.QPointF(13.0, 12.0), QtCore.QPointF(8.0, 17.0))
            # arrow
            self._arrow = None
            self.setArrow(int(collapsed))

        def setArrow(self, arrow_direction: int):
            """
            Sets the direction of the arrow.
            :param arrow_direction: the direction of the arrow (0 = horizontal, 1 = vertical).
            """
            if arrow_direction:
                self._arrow = self._arrow_vertical
            else:
                self._arrow = self._arrow_horizontal
            self.repaint()

        def paintEvent(self, event):
            painter = QPainter(self)
            painter.setBrush(QColor(192, 192, 192))
            painter.setPen(QColor(64, 64, 64))
            painter.drawPolygon(*self._arrow)



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a hex string.

    Example:

        Input:
            6162636465666768696a6b6c6d6e6f70717273747576777 \\
            8797a0a5ec2b02122c2a72425262f28293d3fc2b4603c3e \\
            7c202c2e2d3b3a5f232b272a7e0a30313233343536373839

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HEX (str)', "Thomas Engel", [], context)

    def run(self, text):
        return self._run_lines(text, lambda text_part: bytes.fromhex(text_part).decode('ascii'))

    def can_decode_input(self, input):
        if len(input) % 2 == 0:
            hex = re.search(r'^[a-fA-F0-9]+$', input)
            if hex:
                return True
        return False


from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    Minifies HTML Code.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HTML Minify', "Thomas Engel", ["css-html-js-minify"], context)

    def run(self, text):
        from css_html_js_minify import html_minify
        return html_minify(text)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5 import QtCore
from PyQt5.QtCore import Qt, QPoint, QEvent
from PyQt5.QtGui import QKeySequence, QCursor
from PyQt5.QtWidgets import QLabel, QRadioButton, QCheckBox, QLineEdit, QDialog, QDialogButtonBox, QFormLayout, \
    QGroupBox, QShortcut, QVBoxLayout, QFrame, QHBoxLayout, QPlainTextEdit, QAction

from dpp.core.plugin import PluginConfig


class PluginConfigDialog(QDialog):

    class WidgetWrapper:
        """ Wraps a widget (e.g. QLineEdit, QCheckbox, etc.) and enhances it with a validate and onChange method. """

        def __init__(self, option: PluginConfig.Option.Base, widget, get_value_callback, config, on_change_signal):
            self.name = option.name
            self.widget = widget
            self.config = config
            self._get_value_callback = get_value_callback
            self.validate = lambda codec, input: self.config.validate(option, codec, input)
            self.onChange = on_change_signal

        def _get_value(self):
            return self._get_value_callback()

        value = property(_get_value)

    def __init__(self, context, config, title, codec, icon=None):
        super(PluginConfigDialog, self).__init__()
        self._context = context
        self.config = config
        self._codec = codec
        self._input = None
        self._widgets = {}
        self.setLayout(self._init_main_layout())
        self._build()
        self._init_shortcuts()
        self._validate()
        self.setWindowTitle(title)
        if icon:
            self.setWindowIcon(icon)

    def _init_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.addWidget(self._init_input_frame())
        main_layout.addWidget(self._init_error_frame())
        self._btn_box = self._init_button_box()
        main_layout.addWidget(self._btn_box)
        return main_layout

    def _init_error_frame(self):
        self._error_frame = QFrame()
        layout = QVBoxLayout()
        self._error_text = QLabel("")
        self._error_text.setStyleSheet('QLabel { color: red }')
        layout.addWidget(self._error_text)
        self._error_frame.setLayout(layout)
        self._error_frame.setHidden(True)
        return self._error_frame

    def _init_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        return button_box

    def _init_shortcuts(self):
        def _accept(self):
            if self._btn_box(QDialogButtonBox.Ok).isEnabled():
                self.accept()

        ctrl_return_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Return), self)
        ctrl_return_shortcut.activated.connect(_accept)
        alt_return_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_Return), self)
        alt_return_shortcut.activated.connect(_accept)
        alt_o_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_O), self)
        alt_o_shortcut.activated.connect(_accept)

    def _init_input_frame(self):
        input_frame = QGroupBox()
        self._input_frame_layout = QFormLayout()
        input_frame.setLayout(self._input_frame_layout)
        return input_frame

    def _on_change(self):
        """ Update and validate options. """
        self.config.update({key: self._widgets[key].value for key in self.config.keys()})
        self._btn_box.button(QDialogButtonBox.Ok).setEnabled(self._validate())

    def _validate(self):
        """ Validate the current settings. """
        self._reset_errors()
        for key, widget_wrapper in self._widgets.items():
            message = widget_wrapper.validate(self._codec, self._input)
            if message is not True:
                self._show_error_indicator(widget_wrapper)
                self._show_error_message(message)
                return False

        message = self._try_codec()
        if message:
            self._show_error_message(message)
            return False

        return True

    def _try_codec(self):
        """ Tries to run the codec with the specified input. Either returns None or an error message."""
        try:
            self._codec(self.config, self._input)
        except BaseException as e:
            return str(e)

    def _reset_errors(self):
        """ Hides error message box and resets any error indicators. """
        self._error_frame.setHidden(True)
        self._error_text.setText("")
        for name in self.config.keys():
            self._reset_error(self._widgets[name])

    def _reset_error(self, widget_wrapper):
        """
        Resets any error indication on a specific widget.
        This method can be overwritten to allow customizing visualization of errors for specific widgets.
        """
        widget = widget_wrapper.widget
        if isinstance(widget, QLineEdit):
            widget.setStyleSheet('QLineEdit { }')

    def _show_error_indicator(self, widget_wrapper):
        """
        Indicates an error on a specific widget.
        This method can be overwritten to allow customizing visualization of errors for specific widgets.
        """
        widget = widget_wrapper.widget
        if isinstance(widget, QLineEdit):
            widget.setStyleSheet('QLineEdit { color: red }')

    def _show_error_message(self, message):
        """ Shows an error message within a error-box. """
        self._error_frame.setHidden(False)
        self._error_text.setText(message)

    def _build_option(self, option: PluginConfig.Option.Base) -> WidgetWrapper:
        """
        Automatically build the widget for this option.
        This method can be overwritten to allow building custom widgets.
        :return the widget for this option.
        """
        if isinstance(option, PluginConfig.Option.String):
            x = QLineEdit(option.value)
            w = PluginConfigDialog.WidgetWrapper(option, x, x.text, self.config, x.textChanged)
            w.onChange.connect(lambda evt: self._on_change())
        elif isinstance(option, PluginConfig.Option.Integer):
            x = QLineEdit(option.value)
            w = PluginConfigDialog.WidgetWrapper(option, x, x.text, self.config, x.textChanged)
            w.onChange.connect(lambda evt: self._on_change())
        elif isinstance(option, PluginConfig.Option.Group):
            x = QRadioButton(option.name)
            x.setChecked(option.value)
            w = PluginConfigDialog.WidgetWrapper(option, x, x.isChecked, self.config, x.clicked)
            w.onChange.connect(lambda evt: self._on_change())
        elif isinstance(option, PluginConfig.Option.Boolean):
            x = QCheckBox(option.name)
            x.setChecked(option.value)
            w = PluginConfigDialog.WidgetWrapper(option, x, x.isChecked, self.config, x.clicked)
            w.onChange.connect(lambda evt: self._on_change())
        else:
            raise Exception("Invalid option of type {} detected!".format(type(option)))
        return w

    def _add_option_widget(self, label: QLabel, widget):
        self._input_frame_layout.addRow(label, widget)

    def _build(self):
        """
        Automatically build the widgets for all options.
        This method can be overwritten to allow building custom widgets.
        """
        for key, option in self.config.items():
            label = QLabel()
            if not isinstance(option, PluginConfig.Option.Boolean):
                label.setText(option.name)
            self._widgets[key] = self._build_option(option)
            self._add_option_widget(label, self._widgets[key].widget)

    def setInput(self, text):
        self._input = text
        self._on_change()


class PluginConfigPreviewDialog(PluginConfigDialog):

    def __init__(self, context, config, title, codec, icon=None):
        super(PluginConfigPreviewDialog, self).__init__(context, config, title, codec, icon)

    def _init_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.addWidget(self._init_input_frame())
        main_layout.addWidget(self._init_preview_frame())
        main_layout.addWidget(self._init_error_frame())
        self._btn_box = self._init_button_box()
        main_layout.addWidget(self._btn_box)
        return main_layout

    def _init_preview_frame(self):
        view_frame = QFrame(self)
        view_frame_layout = QHBoxLayout()
        self._txt_preview = QPlainTextEdit(self)
        self._txt_preview.setReadOnly(True)
        self._txt_preview.setLineWrapMode(QPlainTextEdit.NoWrap)
        self._txt_preview.customContextMenuRequested.connect(self._show_preview_frame_context_menu)
        self._txt_preview.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        view_frame_layout.addWidget(self._txt_preview)
        view_frame.setLayout(view_frame_layout)
        return view_frame

    def _init_input_frame(self):
        input_frame = QGroupBox()
        self._input_frame_layout = QHBoxLayout()
        input_frame.setLayout(self._input_frame_layout)
        return input_frame

    def _add_option_widget(self, label: QLabel, widget):
        self._input_frame_layout.addWidget(label)
        self._input_frame_layout.addWidget(widget)

    def _on_change(self):
        super(PluginConfigPreviewDialog, self)._on_change()
        if self._validate():
            self._do_preview()

    def _do_preview(self):
        try:
            result = self._codec(self.config, self._input)
            self._txt_preview.setPlainText(result)
            return True
        except BaseException as e:
            return False

    def _show_preview_frame_context_menu(self, point: QPoint=None):
        """ Displays a customized context menu for the plain view. """
        def _on_plain_text_context_menu_wrap_lines(e: QEvent):
            """ Un-/wraps lines when user clicks the wrap-lines action within the plain views context-menu. """
            if self._txt_preview.lineWrapMode() == QPlainTextEdit.NoWrap:
                self._txt_preview.setLineWrapMode(QPlainTextEdit.WidgetWidth)
            else:
                self._txt_preview.setLineWrapMode(QPlainTextEdit.NoWrap)

        if not point:
            point = QCursor.pos()
        context_menu = self._txt_preview.createStandardContextMenu()
        context_menu.addSeparator()
        wrap_lines_action = QAction(self)
        wrap_lines_action.setText("Wrap Lines")
        wrap_lines_action.setCheckable(True)
        wrap_lines_action.setChecked(self._txt_preview.lineWrapMode() == QPlainTextEdit.WidgetWidth)
        wrap_lines_action.triggered.connect(_on_plain_text_context_menu_wrap_lines)
        context_menu.addAction(wrap_lines_action)
        context_menu.exec(self._txt_preview.mapToGlobal(point))


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtCore import QObject, pyqtSignal


class Listener(QObject):
    """ A set of global signals to emit or connect to. """

    # Signals that a new tab should be created with the specified text
    newTabRequested = pyqtSignal(str)  # text

    # Signals that the selected frame changed (e.g. to update the hex view)
    selectedFrameChanged = pyqtSignal(str, str, str)  # tab_id, frame_id, text

    # Signals that the text inside the codec frame changed (e.g. to update the hex view)
    textChanged = pyqtSignal(str, str, str)  # tab_id, frame_id, text

    # Signals that the text selection inside the codec frame changed (e.g. to update the hex view)
    textSelectionChanged = pyqtSignal(str, str, str)  # tab_id, frame_id, text

    # Signals that text of a codec frame should be changed to the specified text (e.g. when hex view was edited by user)
    textSubmitted = pyqtSignal(str, str, str) # tab_id, frame_id, text

    def __init__(self, context: 'core.context.Context'):
        super(__class__, self).__init__()
        # Logs each event when being triggered
        self.newTabRequested.connect(lambda text:
            context.logger().debug("newTabRequested({})".format(text)))
        self.selectedFrameChanged.connect(lambda tab_id, frame_id, text:
            context.logger().debug("selectedFrameChanged({}, {}, {})".format(tab_id, frame_id, text)))
        self.textChanged.connect(lambda tab_id, frame_id, text:
            context.logger().debug("textChanged({}, {}, {})".format(tab_id, frame_id, text)))
        self.textSelectionChanged.connect(lambda tab_id, frame_id, text:
            context.logger().debug("textSelectionChanged({}, {}, {})".format(tab_id, frame_id, text)))
        self.textSubmitted.connect(lambda tab_id, frame_id, text:
            context.logger().debug("textSubmitted({}, {}, {})".format(tab_id, frame_id, text)))



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a text using Base64.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            0ECJPC% CC3DVED5QDO$D/3EHFE QEA%ET4F3GF:D1PROM84GROSP4A$4L35JX7TROL7CL+7134V$5.L7A1CMK5XG5/C1*96DL6WW66:6C1
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE45', "Thomas Engel", ["base45"], context)

    def run(self, text):
        import base45
        return base45.b45encode(text.encode('utf-8', errors="surrogateescape")).decode('utf-8', errors="surrogateescape")



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QIcon, QKeySequence
from PyQt5.QtWidgets import QDialog, QLabel, QVBoxLayout
from dpp.ui.widget.icon_label import IconLabel


class KeyboardShortcutDialog(QDialog):
    """ A dialog to let users change a keyboard shortcut. """

    __INFO_MESSAGE_DEFAULT_TEXT = "Press Esc to cancel or Backspace to reset the shortcut."
    __INFO_MESSAGE_HINT_STYLE = "QLabel { color: gray }"
    __INFO_MESSAGE_ERROR_STYLE = "QLabel { color: red }"

    def __init__(self, parent, context: 'core.context.Context', keyboard_shortcut_name):
        """
        Initializes the shortcut dialog.
        :param parent: the widget which is calling the dialog.
        :param context: the context of the application
        :param keyboard_shortcut_name: the name of the keyboard shortcut to be changed (e.g. "New Tab").
        """
        super().__init__(parent)
        self._context = context
        self._keyboard_shortcut_name = keyboard_shortcut_name
        self._should_be_reset = False

        layout = QVBoxLayout()
        lbl_title = QLabel("Enter the new keyboard shortcut to change <b>{}</b>".format(keyboard_shortcut_name))
        layout.addWidget(lbl_title)
        layout.setAlignment(lbl_title, Qt.AlignHCenter)

        keyboard_image_path = os.path.join(self._context.getAppPath(), 'images', 'keyboard.png')
        lbl_keyboard_image = IconLabel(self, QIcon(keyboard_image_path))
        lbl_keyboard_image.setFixedSize(QSize(128, 128))
        layout.addWidget(lbl_keyboard_image)
        layout.setAlignment(lbl_keyboard_image, Qt.AlignHCenter)

        self._lbl_keyboard_shortcut = QLabel()
        layout.addWidget(self._lbl_keyboard_shortcut)
        layout.setAlignment(self._lbl_keyboard_shortcut, Qt.AlignHCenter)

        self._lbl_hint = QLabel(self.__INFO_MESSAGE_DEFAULT_TEXT)
        self._lbl_hint.setStyleSheet(self.__INFO_MESSAGE_HINT_STYLE)
        layout.setAlignment(Qt.AlignCenter)
        layout.setAlignment(self._lbl_hint, Qt.AlignHCenter)
        layout.addWidget(self._lbl_hint)

        layout.setContentsMargins(10, 10, 10, 10)
        self.setLayout(layout)
        self.setMinimumWidth(150)
        self.setWindowTitle("Set Keyboard Shortcut")

    def _show_error_message(self, text):
        self._lbl_hint.setStyleSheet(self.__INFO_MESSAGE_ERROR_STYLE)
        self._lbl_hint.setText(text)

    def _show_info_message(self, text=""):
        if not text:
            text = self.__INFO_MESSAGE_DEFAULT_TEXT
        self._lbl_hint.setStyleSheet(self.__INFO_MESSAGE_HINT_STYLE)
        self._lbl_hint.setText(text)

    def setKeyboardShortcut(self, shortcut: str):
        self._lbl_keyboard_shortcut.setText(shortcut)

    def keyboardShortcut(self) -> str:
        return self._lbl_keyboard_shortcut.text()

    def shouldBeReset(self):
        return self._should_be_reset

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.setKeyboardShortcut("")
            self.close()
            return

        if event.key() == Qt.Key_Backspace:
            self.setKeyboardShortcut("")
            self._should_be_reset = True
            self.close()
            return

        try:
            key_sequence = QKeySequence(event.modifiers() | event.key())
            key_sequence_ascii_representation = key_sequence.toString().encode('ascii',
                                                                               errors='ignore').decode()
            self.setKeyboardShortcut(key_sequence_ascii_representation)
        except Exception as e:
            # Ignoring unknown errors since we can't do anything about it.
            pass

    def keyReleaseEvent(self, event):
        # Only allow complete key-sequences (e.g. 'A', 'Ctrl+A').
        if len(self.keyboardShortcut()) == 0 or self.keyboardShortcut().endswith("+"):
            # Remove incomplete key-sequences (e.g. 'Ctrl+') when releasing keys.
            self.setKeyboardShortcut("")
            self._show_info_message()
            return

        # Check for duplicates
        for keyboard_shortcut in self._context.getShortcuts():
            if keyboard_shortcut.key() == self.keyboardShortcut() and \
                    keyboard_shortcut.name() != self._keyboard_shortcut_name:
                self._show_error_message("<b>{}</b> is already being used for <b>{}</b>.".format(
                    keyboard_shortcut.key(), keyboard_shortcut.name(remove_anchors=True)))
                return

        self.close()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QKeySequence
from PyQt5.QtWidgets import QVBoxLayout, QFrame, QShortcut, QDialogButtonBox

from dpp.ui import MainWindow, MainWindowWidget, CodecTab


class DecoderPlusPlusWindow(MainWindow):
    """ The DecoderPlusPlus application. """

    def __init__(self, context: 'core.context.Context', input: str=None):
        """
        Initializes the DecoderPlusPlus application.
        :param context: the application context.
        :param input: the user input.
        """
        super().__init__(context, input)
        self._main_window_widget = MainWindowWidget(self, self._context, input)
        self.setCentralWidget(self._main_window_widget)
        self.show()

    def newTab(self, input: str):
        """ Opens a new tab with the specified input as content for the first codec frame. """
        self._main_window_widget.newTab(input)


class DecoderPlusPlusDialog(MainWindow):
    """
    The DecoderPlusPlusDialog with OK- and Cancel button.
    When the OK-button is triggered the transformed text is printed to stdout.
    When the Cancel-button is triggered or the user exist the application in any other way the initial input is
    printed to stdout.
    """

    def __init__(self, context: 'core.context.Context', input: str=None):
        """
        Initializes the DecoderPlusPlus dialog.
        :param context: the application context.
        :param input: the user input.
        """
        super().__init__(context, input)

        # Store the initial user input which is returned when the user cancels application.
        self._user_input = input

        # As long as the "OK"-button is not triggered we assume that the application was closed by the user.
        self._application_was_canceled = True

        # Build main window
        self._main_window_widget = QFrame()
        self._main_window_widget.setLayout(QVBoxLayout())
        self._codec_tab_widget = CodecTab(self, self._context, context.plugins())
        self._codec_tab_widget.frames().getFocusedFrame().setInputText(input)
        self._main_window_widget.layout().addWidget(self._codec_tab_widget)
        self._main_window_widget.layout().addWidget(self._init_button_box())
        self.setCentralWidget(self._main_window_widget)

        # Setup additional shortcuts to allow user to quickly hit the accept button.
        self._setup_shortcuts()

        # Do not show statusbar in dialog mode.
        self.statusBar().hide()

        # Show dialog.
        self.show()

    def _setup_shortcuts(self):
        """ Setup shortcuts to allow user to quickly hit the accept button. """
        ctrl_return_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Return), self)
        ctrl_return_shortcut.activated.connect(self.onAccept)
        alt_return_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_Return), self)
        alt_return_shortcut.activated.connect(self.onAccept)
        alt_o_shortcut = QShortcut(QKeySequence(Qt.ALT + Qt.Key_O), self)
        alt_o_shortcut.activated.connect(self.onAccept)

    def _init_button_box(self):
        """ Initialize the central dialog buttons.  """
        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self.onAccept)
        button_box.rejected.connect(self.onReject)
        button_box.setContentsMargins(0, 35, 0, 0)
        return button_box

    def onAccept(self):
        """ """
        self._application_was_canceled = False
        self.close()

    def onReject(self):
        self.close()

    def closeEvent(self, event):
        """ Handles the closeEvent which is triggered when the user exits the application. """
        if self._application_was_canceled:
            # Return the initial user input (here: no change) when user cancelled application.
            print(self._user_input, end='')
        else:
            # Return the transformed input when user triggered the OK-button.
            codec_frames = self._codec_tab_widget.getFrames()
            print(codec_frames[-1].getInputText(), end = '')
        super().closeEvent(event)



from dpp.core.plugin import ScriptPlugin
from os import linesep
from urlextract import URLExtract


class Plugin(ScriptPlugin):
    """
    Extracts URLs from the input string.
    """

    extractor = URLExtract()

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Extract URLs', "Thomas Engel", ['urlextract'], context)

    def run(self, text):
        return linesep.join(self.extractor.find_urls(text))



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect

from PyQt5 import QtCore
from PyQt5.QtWidgets import QVBoxLayout, QGroupBox, QLabel, QFrame

from dpp.core.plugin import AbstractPlugin


class PluginFrame(QFrame):

    def __init__(self, context, plugin: AbstractPlugin, parent=None):
        super(PluginFrame, self).__init__(parent)

        self._context = context
        self._plugin = plugin

        layout = QVBoxLayout()
        layout.addWidget(self._make_group_text_field("Name", plugin.name()))
        layout.addWidget(self._make_group_text_field("Description", self._clean_description(plugin.__doc__)))
        if plugin.dependencies():
            layout.addWidget(self._make_group_text_field("Dependencies", self._clean_dependencies(plugin.dependencies())))
        layout.addWidget(self._make_group_text_field("Author", plugin.author()))
        layout.addStretch(1)
        self.setLayout(layout)

    def _make_group_text_field(self, title, text):
        grp_field = QGroupBox(title)
        grp_field_layout = QVBoxLayout()
        txt_field = QLabel()
        txt_field.setTextFormat(QtCore.Qt.RichText)
        txt_field.setWordWrap(True)
        txt_field.setText(text)
        grp_field_layout.addWidget(txt_field)
        grp_field.setLayout(grp_field_layout)
        return grp_field

    def _clean_description(self, description_text):
        return "<br/>".join(inspect.cleandoc(description_text).split("\n"))

    def _clean_dependencies(self, dependencies):
        dependencies_text = ""
        for dependency in dependencies:
            is_unresolved_dependency = not self._plugin.check_dependency(dependency)
            if is_unresolved_dependency:
                dependencies_text += "<div style='color:red;'>{} (unresolved)</div>".format(dependency)
            else:
                dependencies_text += "<div>{}</div>".format(dependency)
            dependencies_text += "</br>"
        return dependencies_text



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA256.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            0a4035197aa3b94d8ee2ff7d5b286636 \\
            f6264f6c96ffccf3c4b777a8fb9be674
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA256', "Thomas Engel", ["hashlib"], context)

    def run(self, text):
        import hashlib
        return hashlib.sha256(text.encode('utf-8', errors='surrogateescape')).hexdigest()



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5 import QtCore
from PyQt5.QtCore import QSize, QSortFilterProxyModel, pyqtSignal
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QListView


class ListWidget(QListView):
    """ Enhanced QListWidget with additional methods for selecting items by name or index. """

    keyPressed = pyqtSignal("PyQt_PyObject")

    def __init__(self):
        # TODO: Shouldnt we use super here?
        QListView.__init__(self)
        proxy = QSortFilterProxyModel(self)
        proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self._model = QStandardItemModel(self)
        proxy.setSourceModel(self._model)
        self.setModel(proxy)

    def select_item_by_text(self, text):
        """ Highligts and selects an item matching the text in the list-widget. """
        if text:
            for index in range(self.count()):
                if str(self.item(index).text()).startswith(text):
                    self.select_item_by_index(index)
                    self.setFocus()
                    break

    def select_item_by_index(self, index):
        """ Highlights and selects an item for the purpose of navigation (e.g. arrow-up, arrow-down). """
        self.setCurrentIndex(self._model.index(index, 0))

    def keyPressEvent(self, evt):
        key = evt.key()
        if key == QtCore.Qt.Key_Down or key == QtCore.Qt.Key_Up:
            # Select item above/below in list view.
            super(ListWidget, self).keyPressEvent(evt)
        elif key == QtCore.Qt.Key_Enter or key == QtCore.Qt.Key_Return:
            # Usually presses ok button and exits dialog.
            super(ListWidget, self).keyPressEvent(evt)
        else:
            # Signal all other events to listeners.
            self.keyPressed.emit(evt)

    def sizeHint(self):
        """ Set a reasonable size for the list widget. """
        s = QSize()
        s.setHeight(super(ListWidget, self).sizeHint().height())
        s.setWidth(self.sizeHintForColumn(0))
        return s

    def setFilterCaseSensitive(self, cs: int):
        """
        Sets the case-sensitivity of the proxy-controller.
        :param cs: the case-sensitivity (QtCore.Qt.CaseSensitive, QtCore.Qt.CaseInsensitive).
        """
        self.model().setFilterCaseSensitivity(cs)

    def addItem(self, item: QStandardItem):
        """ Adds an item to the list. """
        self._model.appendRow(item)

    def item(self, index: int):
        """
        Returns the item at the specified index.
        :raises exception when the index is out of range.
        """
        return self._model.item(index)



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using FreeBSD NT.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            $3$$d6a56df12312ac8f5e208d75164cdcb2
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('FreeBSD NT', "Thomas Engel", ["passlib"], context)

    def run(self, text):
        from passlib.hash import bsd_nthash
        return bsd_nthash.encrypt(text.encode('utf-8', errors='surrogateescape'))



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QPoint
from PyQt5.QtWidgets import QTabBar, QLineEdit, QWidget


class TabBar(QTabBar):
    """QTabBar with double click signal and tab rename behavior."""

    # tabRenamed(index, old_name, new_name)
    tabRenamed = pyqtSignal('PyQt_PyObject', 'PyQt_PyObject', 'PyQt_PyObject')
    tabDoubleClicked = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.__edit_mode_activated = False
        self.setMovable(True)
        self.installEventFilter(self)

    def eventFilter(self, source, event):
        """ Allow moving tabs under certain conditions. """

        if event.type() == QtCore.QEvent.Type.MouseMove:

            if source.currentIndex() == self.count() - 1:
                # Block MouseMove for last tab.
                return True
            else:
                last_tab = self.tabRect(self.count() - 1)
                if event.pos().x() > last_tab.x():
                    # Block movement beyond last tab.
                    return True

                # Retrieve pressed tab.
                # Do not use y-position of event.pos() since mouse may not focus tab all the time.
                pressed_tab_pos = QPoint(event.pos().x(), self.y())
                pressed_tab = self.tabRect(self.tabAt(pressed_tab_pos))
                if (pressed_tab.x() + pressed_tab.width()) >= last_tab.x():
                    # Block movement of pressed tab when reaching beginning of last tab.
                    return True

        return QWidget.eventFilter(self, source, event)

    def mouseDoubleClickEvent(self, event):
        if not self.__edit_mode_activated:
            self.__edit_mode_activated = True
            tab_index = self.tabAt(event.pos())
            self.tabDoubleClicked.emit(tab_index)
            self.start_rename(tab_index)

    def renameTab(self, index=None):
        if not index:
            index = self.currentIndex()
        self.start_rename(index)

    def start_rename(self, tab_index):
        if not self.isTabEnabled(tab_index):
            return
        self.__edited_tab = tab_index
        rect = self.tabRect(tab_index)
        top_margin = 3
        left_margin = 6
        self.__edit = QLineEdit(self)
        self.__edit.show()
        self.__edit.move(rect.left() + left_margin, rect.top() + top_margin)
        self.__edit.resize(rect.width() - 2 * left_margin, rect.height() - 2 * top_margin)
        self.__edit.setText(self.tabText(tab_index))
        self.__edit.selectAll()
        self.__edit.setFocus()
        self.__edit.editingFinished.connect(self.finish_rename)

    @pyqtSlot()
    def finish_rename(self):
        old_name = self.tabText(self.__edited_tab)
        new_name = self.__edit.text()
        self.setTabText(self.__edited_tab, self.__edit.text())
        self.__edit.deleteLater()
        self.__edit_mode_activated = False
        self.tabRenamed.emit(self.__edited_tab, old_name, new_name)




from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using SHA3 384.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            4fa54dbfbdde8c191a64879a1bed6a062da45f53c82e7eb0fc59362e32f506cded9d17e69f881231b16395a259595c6b
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('SHA3 384', "Thomas Engel", ["_pysha3"], context)

    def run(self, text):
        import _pysha3
        return _pysha3.sha3_384(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a string to a hex string.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            6162636465666768696a6b6c6d6e6f70717273747576777 \\
            8797a0a5ec2b02122c2a72425262f28293d3fc2b4603c3e \\
            7c202c2e2d3b3a5f232b272a7e0a30313233343536373839
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('HEX (str)', "Thomas Engel", ["codecs"], context)

    def run(self, text):
        import codecs
        return codecs.encode(text.encode('utf-8', errors='surrogateescape'), 'hex').decode('utf-8', errors='surrogateescape')



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using Adler-32.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            0xc713178c
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Adler-32', "Thomas Engel", ["zlib"], context)

    def run(self, text):
        import zlib
        return hex(zlib.adler32(text.encode('utf-8', errors='surrogateescape')))



from dpp.core.plugin import ScriptPlugin


class Plugin(ScriptPlugin):
    """
    Escapes the input string.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Escape String', "Thomas Engel", [], context)

    def run(self, text):
        return repr(text)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import qtawesome
from typing import List

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QPushButton, QMenu, QAction

from dpp.core.plugin import DecoderPlugin, IdentifyPlugin, AbstractPlugin


class IdentifyFormatButton(QFrame):
    """ A button which provides a identify-format functionality. """

    def __init__(self, parent, plugins: List[AbstractPlugin], get_input_callback, select_plugin_callback, logger):
        super(__class__, self).__init__(parent)
        self._plugins = plugins
        self._get_input = get_input_callback
        self._select_plugin = select_plugin_callback
        self._logger = logger
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 6, 0, 0)
        self._button = self._init_button()
        layout.addWidget(self._button)
        self.setLayout(layout)

    def _init_button(self):
        button = QPushButton("Identify format")
        menu = QMenu(self)
        menu.aboutToShow.connect(lambda: self._populate_button_menu(self._get_input()))
        button.setMenu(menu)
        return button

    def _can_plugin_identify_input(self, plugin: AbstractPlugin, input: str) -> bool:
        """ Returns whether the plugin can identify the specified input. """
        try:
            # Check whether it is a decoder plugin and it thinks it can decode the input
            if isinstance(plugin, DecoderPlugin) and not plugin.can_decode_input(input):
                return False
            # Check whether decoder/identifier actually can process the input without any error
            plugin.select(input)
            return True
        except:
            return False

    def _get_matching_plugins(self, input) -> List[AbstractPlugin]:
        """ Returns a list of matching plugins for the specified input. """
        if not input:
            return []

        return [plugin for plugin in self._plugins if self._can_plugin_identify_input(plugin, input)]

    def _populate_button_menu(self, input):
        """ Populates the button menu with a list of matching plugins for the specified input. """
        menu = self._button.menu()
        menu.clear()
        plugins = self._get_matching_plugins(input)
        if not plugins:
            action = menu.addAction("No matching format found ...")
            action.setEnabled(False)
            return

        actions = []
        for plugin in plugins:
            if isinstance(plugin, DecoderPlugin):
                action = QAction(qtawesome.icon("ei.indent-left"), plugin.name(), self)
                action.triggered.connect(lambda chk, item=plugin: self._select_plugin(item))
                actions.append(action)
            elif isinstance(plugin, IdentifyPlugin):
                for identifier in plugin.run(input).splitlines():
                    action = QAction(qtawesome.icon("ei.align-justify"), identifier, self)
                    actions.append(action)

        menu.addActions(sorted(actions, key=lambda action: action.text()))



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a Base54 string.

    Example:

        Input:
            0ECJPC% CC3DVED5QDO$D/3EHFE QEA%ET4F3GF:D1PROM84GROSP4A$4L35JX7TROL7CL+7134V$5.L7A1CMK5XG5/C1*96DL6WW66:6C1

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE45', "Thomas Engel", ["base45"], context)

    def run(self, text):
        import base45
        return base45.b45decode(str.encode(text)).decode('utf-8', errors="surrogateescape")

    def can_decode_input(self, input):
        if input.startswith("0o") and input[2:].isdigit():
            # This looks more like an octal encoding.
            return False
        if re.search(r'^[A-Z0-9 $%*+-./:]*$', input):
            return True
        return False




import json
from json import JSONDecodeError
import os

import qtawesome

from PyQt5.QtWidgets import QDialog

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigPreviewDialog


class Plugin(ScriptPlugin):
	"""
	Opens a dialog to filter xml text by certain jq expression.

	Example:

		Input:
			{"a": [{"b": "c"}, {"b": "d"}]}

		Expression:
			.a[].b

		Output:
			"c"
			"d"

	"""

	class Option(object):
		Expression = PluginConfig.Option.Label("expression", "Expression:")

	class Codec:

		def run(self, config: PluginConfig, text: str):
			try:
				import pyjq as pyjq
				expression = config.get(Plugin.Option.Expression).value
				return os.linesep.join([
					json.dumps(item) for item in pyjq.all(expression, json.loads(text))
				])
			except JSONDecodeError as e:
				raise Exception("Error decoding json!")
			except Exception as e:
				# Ignore exceptions - most likely an error in the jq expression
				pass

	def __init__(self, context):
		# Name, Author, Dependencies
		super().__init__('JQ', "Thomas Engel", ["pyjq"], context)
		self._context = context
		self._logger = context.logger
		self._codec = Plugin.Codec()
		self.config.add(PluginConfig.Option.String(
			label=Plugin.Option.Expression,
			value="",
			description="jq expression to filter by",
			is_required=True
		))
		self._dialog = None
		self._dialog_return_code = QDialog.Accepted

	def select(self, text: str):
		if not self._dialog:
			self._dialog = PluginConfigPreviewDialog(self._context, self.config.clone(), "jq",
													 self._codec.run, qtawesome.icon("fa.filter"))

		self._dialog.setInput(text)
		self._dialog_return_code = self._dialog.exec_()

		if self._dialog_return_code != QDialog.Accepted:
			# User clicked the Cancel-Button.
			raise AbortedException("Aborted")

		self.config.update(self._dialog.config)
		return self.run(text)

	def title(self):
		return "Filter by jq expression '{}'".format(self.config.get(Plugin.Option.Expression).value)

	def run(self, text: str):
		return self._codec.run(self.config, text)



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Encodes a string to an URL. Spaces are encoded to plus-signs.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            abcdefghijklmnopqrstuvwxyz \\
            %0A%5E%C2%B0%21%22%C2%A7%24%25%26/%28%29%3D%3F%C2%B4%60%3C%3E%7C+%2C.-%3B%3A_%23%2B%27%2A%7E%0A \\
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('URL+', "Thomas Engel", ["urllib"], context)

    def run(self, text):
        import urllib.parse
        return urllib.parse.quote_plus(text.encode('utf-8', errors='surrogateescape'))



from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using MD2.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            f5299f20b7bf89a89bc575f1dcea358a
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('MD2', "Thomas Engel", ["Crypto"], context)

    def run(self, text):
        from Crypto.Hash import MD2
        return MD2.new(text.encode('utf-8', errors='surrogateescape')).hexdigest()



from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a string of comma separated decimals to its character representation.

    Example:

        Input:
            97,98,99,100,101,102,103,104,\\
            105,106,107,108,109,110,111,112,\\
            ...
            10,48,49,50,51,52,53,54,\\
            55,56,57

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('DEC (str)', "Thomas Engel", [], context)

    def run(self, text):
        return "".join(chr(int(c)) for c in text.split(","))



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
        Encodes a string using ZLib.

        Example:

            Input:
                abcdefghijklmnopqrstuvwxyz
                ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
                0123456789

            Output:
                [bytes]
        """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Zlib', "Thomas Engel", ["zlib"], context)

    def run(self, text):
        import zlib
        return zlib.compress(text.encode('utf-8', errors='surrogateescape')).decode('utf-8', errors='surrogateescape')
